"""
***************************************************************************
    RandomPointsLayer.py
    ---------------------
    Date                 : April 2014
    Copyright            : (C) 2014 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "April 2014"
__copyright__ = "(C) 2014, Alexander Bruy"

import os
import random

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsPointXY,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterDistance,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterNumber,
    QgsSpatialIndex,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm
from processing.tools import vector

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class RandomPointsLayer(QgisAlgorithm):
    INPUT = "INPUT"
    POINTS_NUMBER = "POINTS_NUMBER"
    MIN_DISTANCE = "MIN_DISTANCE"
    OUTPUT = "OUTPUT"

    def icon(self):
        return QgsApplication.getThemeIcon(
            "/algorithms/mAlgorithmRandomPointsWithinExtent.svg"
        )

    def svgIconPath(self):
        return QgsApplication.iconPath(
            "/algorithms/mAlgorithmRandomPointsWithinExtent.svg"
        )

    def group(self):
        return self.tr("Vector creation")

    def groupId(self):
        return "vectorcreation"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input layer"),
                [QgsProcessing.SourceType.TypeVectorPolygon],
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.POINTS_NUMBER,
                self.tr("Number of points"),
                QgsProcessingParameterNumber.Type.Integer,
                1,
                False,
                1,
                1000000000,
            )
        )
        self.addParameter(
            QgsProcessingParameterDistance(
                self.MIN_DISTANCE,
                self.tr("Minimum distance between points"),
                0,
                self.INPUT,
                False,
                0,
                1000000000,
            )
        )
        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Random points"),
                type=QgsProcessing.SourceType.TypeVectorPoint,
            )
        )

    def name(self):
        return "randompointsinlayerbounds"

    def displayName(self):
        return self.tr("Random points in layer bounds")

    def shortDescription(self):
        return self.tr(
            "Creates a new point layer with a given number of random points, "
            "all of them within the extent of a given layer."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a new point layer with a given number of random points, "
            "all of them within the extent of a given layer. A distance factor can be specified, "
            "to avoid points being too close to each other."
        )

    def documentationFlags(self):
        return Qgis.ProcessingAlgorithmDocumentationFlag.RegeneratesPrimaryKey

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        pointCount = self.parameterAsDouble(parameters, self.POINTS_NUMBER, context)
        minDistance = self.parameterAsDouble(parameters, self.MIN_DISTANCE, context)

        bbox = source.sourceExtent()
        sourceIndex = QgsSpatialIndex(source, feedback)

        fields = QgsFields()
        fields.append(QgsField("id", QMetaType.Type.Int, "", 10, 0))

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            fields,
            QgsWkbTypes.Type.Point,
            source.sourceCrs(),
            QgsFeatureSink.SinkFlag.RegeneratePrimaryKey,
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        nPoints = 0
        nIterations = 0
        maxIterations = pointCount * 200
        total = 100.0 / pointCount if pointCount else 1

        index = QgsSpatialIndex()
        points = {}

        random.seed()

        while nIterations < maxIterations and nPoints < pointCount:
            if feedback.isCanceled():
                break

            rx = bbox.xMinimum() + bbox.width() * random.random()
            ry = bbox.yMinimum() + bbox.height() * random.random()

            p = QgsPointXY(rx, ry)
            geom = QgsGeometry.fromPointXY(p)
            ids = sourceIndex.intersects(geom.buffer(5, 5).boundingBox())
            if len(ids) > 0 and vector.checkMinDistance(p, index, minDistance, points):
                request = (
                    QgsFeatureRequest().setFilterFids(ids).setSubsetOfAttributes([])
                )
                for f in source.getFeatures(request):
                    if feedback.isCanceled():
                        break

                    tmpGeom = f.geometry()
                    if geom.within(tmpGeom):
                        f = QgsFeature(nPoints)
                        f.initAttributes(1)
                        f.setFields(fields)
                        f.setAttribute("id", nPoints)
                        f.setGeometry(geom)
                        sink.addFeature(f, QgsFeatureSink.Flag.FastInsert)
                        feedback.featureAddedToSink(self.OUTPUT)
                        index.addFeature(f)
                        points[nPoints] = p
                        nPoints += 1
                        feedback.setProgress(int(nPoints * total))
            nIterations += 1

        if nPoints < pointCount:
            feedback.pushInfo(
                self.tr(
                    "Could not generate requested number of random points. "
                    "Maximum number of attempts exceeded."
                )
            )

        sink.finalize()
        feedback.featureSinkFinalized(self.OUTPUT)
        return {self.OUTPUT: dest_id}



"""
***************************************************************************
    utils.py
    ---------------------
    Date                 : November 2009
    Copyright            : (C) 2009 by Martin Dobias
    Email                : wonder dot sk at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Martin Dobias"
__date__ = "November 2009"
__copyright__ = "(C) 2009, Martin Dobias"

"""
QGIS utilities module

"""
import builtins
import configparser
import functools
import glob
import os
import os.path
import sys
import time
import traceback
import warnings
from typing import Optional

from qgis.core import (
    Qgis,
    QgsMessageLog,
    QgsMessageOutput,
    QgsSettingsTree,
    qgsfunction,
)
from qgis.gui import QgsMessageBar
from qgis.PyQt.QtCore import (
    QT_VERSION_STR,
    QCoreApplication,
    QLocale,
    QThread,
    QUrl,
    qDebug,
)
from qgis.PyQt.QtGui import QDesktopServices
from qgis.PyQt.QtWidgets import QApplication, QPushButton

builtins.__dict__["unicode"] = str
builtins.__dict__["basestring"] = str
builtins.__dict__["long"] = int
builtins.__dict__["Set"] = set

# ######################
# ERROR HANDLING

warnings.simplefilter("default")
warnings.filterwarnings("ignore", "the sets module is deprecated")


def showWarning(message, category, filename, lineno, file=None, line=None):
    stk = ""
    for s in traceback.format_stack()[:-2]:
        if hasattr(s, "decode"):
            stk += s.decode(sys.getfilesystemencoding())
        else:
            stk += s
    if hasattr(filename, "decode"):
        decoded_filename = filename.decode(sys.getfilesystemencoding())
    else:
        decoded_filename = filename
    QgsMessageLog.logMessage(
        f"warning:{warnings.formatwarning(message, category, decoded_filename, lineno)}\ntraceback:{stk}",
        QCoreApplication.translate("Python", "Python warning"),
    )


def showException(
    type, value, tb, msg, messagebar=False, level=Qgis.MessageLevel.Warning
):
    if msg is None:
        msg = QCoreApplication.translate(
            "Python", "An error has occurred while executing Python code:"
        )

    logmessage = ""
    for s in traceback.format_exception(type, value, tb):
        logmessage += s.decode("utf-8", "replace") if hasattr(s, "decode") else s

    title = QCoreApplication.translate("Python", "Python error")
    QgsMessageLog.logMessage(logmessage, title, level)

    try:
        blockingdialog = QApplication.instance().activeModalWidget()
        window = QApplication.instance().activeWindow()
    except:
        blockingdialog = QApplication.activeModalWidget()
        window = QApplication.activeWindow()

    # Still show the normal blocking dialog in this case for now.
    if blockingdialog or not window or not messagebar or not iface:
        open_stack_dialog(type, value, tb, msg)
        return

    bar = iface.messageBar() if iface else None

    # If it's not the main window see if we can find a message bar to report the error in
    if not window.objectName() == "QgisApp":
        widgets = window.findChildren(QgsMessageBar)
        if widgets:
            # Grab the first message bar for now
            bar = widgets[0]

    item = bar.currentItem()
    if item and item.property("Error") == msg:
        # Return of we already have a message with the same error message
        return

    widget = bar.createMessage(
        title,
        msg
        + " "
        + QCoreApplication.translate(
            "Python", "See message log (Python Error) for more details."
        ),
    )
    widget.setProperty("Error", msg)
    stackbutton = QPushButton(
        QCoreApplication.translate("Python", "Stack trace"),
        pressed=functools.partial(open_stack_dialog, type, value, tb, msg),
    )
    button = QPushButton(
        QCoreApplication.translate("Python", "View message log"),
        pressed=show_message_log,
    )
    widget.layout().addWidget(stackbutton)
    widget.layout().addWidget(button)
    bar.pushWidget(widget, Qgis.MessageLevel.Warning)


def show_message_log(pop_error=True):
    if pop_error:
        iface.messageBar().popWidget()

    iface.openMessageLog()


def open_stack_dialog(type, value, tb, msg, pop_error=True):
    if pop_error and iface is not None:
        iface.messageBar().popWidget()

    if msg is None:
        msg = QCoreApplication.translate(
            "Python", "An error has occurred while executing Python code:"
        )

    # TODO Move this to a template HTML file
    txt = """<font color="red"><b>{msg}</b></font>
<br>
<h3>{main_error}</h3>
<pre>
{error}
</pre>
<br>
<b>{version_label}</b> {num}
<br>
<b>{qgis_label}</b> {qversion} {qgisrelease}, {devversion}
<br>
<h4>{pypath_label}</h4>
<ul>
{pypath}
</ul>"""

    error = ""
    lst = traceback.format_exception(type, value, tb)
    for s in lst:
        error += s.decode("utf-8", "replace") if hasattr(s, "decode") else s
    error = error.replace("\n", "<br>")

    main_error = (
        lst[-1].decode("utf-8", "replace") if hasattr(lst[-1], "decode") else lst[-1]
    )

    version_label = QCoreApplication.translate("Python", "Python version:")
    qgis_label = QCoreApplication.translate("Python", "QGIS version:")
    pypath_label = QCoreApplication.translate("Python", "Python Path:")
    txt = txt.format(
        msg=msg,
        main_error=main_error,
        error=error,
        version_label=version_label,
        num=sys.version,
        qgis_label=qgis_label,
        qversion=Qgis.QGIS_VERSION,
        qgisrelease=Qgis.QGIS_RELEASE_NAME,
        devversion=Qgis.QGIS_DEV_VERSION,
        pypath_label=pypath_label,
        pypath="".join(f"<li>{path}</li>" for path in sys.path),
    )

    txt = txt.replace("  ", "&nbsp; ")  # preserve whitespaces for nicer output

    dlg = QgsMessageOutput.createMessageOutput()
    dlg.setTitle(msg)
    dlg.setMessage(txt, QgsMessageOutput.MessageType.MessageHtml)
    dlg.showMessage()


def qgis_excepthook(type, value, tb):
    # detect if running in the main thread
    in_main_thread = (
        QCoreApplication.instance() is None
        or QThread.currentThread() == QCoreApplication.instance().thread()
    )

    # only use messagebar if running in main thread - otherwise it will crash!
    showException(type, value, tb, None, messagebar=in_main_thread)


def installErrorHook():
    """
    Installs the QGIS application error/warning hook. This causes Python exceptions
    to be intercepted by the QGIS application and shown in the main window message bar
    and in custom dialogs.

    Generally you shouldn't call this method - it's automatically called by
    the QGIS app on startup, and has no use in standalone applications and scripts.
    """
    sys.excepthook = qgis_excepthook
    warnings.showwarning = showWarning


def uninstallErrorHook():
    sys.excepthook = sys.__excepthook__


# initialize 'iface' object
iface = None


def initInterface(pointer):
    from qgis.gui import QgisInterface
    from qgis.PyQt.sip import wrapinstance

    global iface
    iface = wrapinstance(pointer, QgisInterface)


#######################
# PLUGINS

# The current path for home directory Python plugins.
HOME_PLUGIN_PATH: Optional[str] = None

# list of plugin paths. it gets filled in by the QGIS python library
plugin_paths = []

# dictionary of plugins
plugins = {}

plugin_times = {}

# list of active (started) plugins
active_plugins = []

# list of plugins in plugin directory and home plugin directory
available_plugins = []

# dictionary of plugins providing metadata in a text file (metadata.txt)
# key = plugin package name, value = config parser instance
plugins_metadata_parser = {}


def findPlugins(path):
    """for internal use: return list of plugins in given path"""
    for plugin in glob.glob(path + "/*"):
        if not os.path.isdir(plugin):
            continue
        if not os.path.exists(os.path.join(plugin, "__init__.py")):
            continue

        metadataFile = os.path.join(plugin, "metadata.txt")
        if not os.path.exists(metadataFile):
            continue

        cp = configparser.ConfigParser()

        try:
            with open(metadataFile, encoding="utf8") as f:
                cp.read_file(f)
        except:
            cp = None

        pluginName = os.path.basename(plugin)
        yield (pluginName, cp)


def metadataParser() -> dict:
    """Used by other modules to access the local parser object"""
    return plugins_metadata_parser


def updateAvailablePlugins(sort_by_dependencies=False):
    """Go through the plugin_paths list and find out what plugins are available."""
    # merge the lists
    plugins = []
    metadata_parser = {}
    plugin_name_map = {}
    for pluginpath in plugin_paths:
        for plugin_id, parser in findPlugins(pluginpath):
            if parser is None:
                continue
            if plugin_id not in plugins:
                plugins.append(plugin_id)
                metadata_parser[plugin_id] = parser
                plugin_name_map[parser.get("general", "name")] = plugin_id

    global plugins_metadata_parser
    plugins_metadata_parser = metadata_parser

    global available_plugins
    available_plugins = (
        _sortAvailablePlugins(plugins, plugin_name_map)
        if sort_by_dependencies
        else plugins
    )


def _sortAvailablePlugins(
    plugins: list[str], plugin_name_map: dict[str, str]
) -> list[str]:
    """Place dependent plugins after their dependencies

    1. Make a copy of plugins list to modify it.
    2. Get a plugin dependencies dict.
    3. Iterate plugins and leave the real work to _move_plugin()

    :param list plugins: List of available plugin ids
    :param dict plugin_name_map: Map of plugin_names and plugin_ids, because
                                 get_plugin_deps() only returns plugin names
    :return: List of plugins sorted by dependencies.
    """
    sorted_plugins = plugins.copy()
    visited_plugins = []

    deps = {}
    for plugin in plugins:
        deps[plugin] = [plugin_name_map.get(dep, "") for dep in get_plugin_deps(plugin)]

    for plugin in plugins:
        _move_plugin(plugin, deps, visited_plugins, sorted_plugins)

    return sorted_plugins


def _move_plugin(
    plugin: str,
    deps: dict[str, list[str]],
    visited: list[str],
    sorted_plugins: list[str],
):
    """Use recursion to move a plugin after its dependencies in a list of
    sorted plugins.

    Notes:
    This function modifies both visited and sorted_plugins lists.
    This function will not get trapped in circular dependencies. We avoid a
    maximum recursion error by calling return when revisiting a plugin.
    Therefore, if a plugin A depends on B and B depends on A, the order will
    work in one direction (e.g., A depends on B), but the other direction won't
    be satisfied. After all, a circular plugin dependency should not exist.

    :param str plugin: Id of the plugin that should be moved in sorted_plugins.
    :param dict deps: Dictionary of plugin dependencies.
    :param list visited: List of plugins already visited.
    :param list sorted_plugins: List of plugins to be modified and sorted.
    """
    if plugin in visited:
        return
    elif plugin not in deps or not deps[plugin]:
        visited.append(plugin)  # Plugin with no dependencies
    else:
        visited.append(plugin)

        # First move dependencies
        for dep in deps[plugin]:
            _move_plugin(dep, deps, visited, sorted_plugins)

        # Remove current plugin from sorted
        # list to get dependency indices
        max_index = sorted_plugins.index(plugin)
        sorted_plugins.pop(max_index)

        for dep in deps[plugin]:
            idx = sorted_plugins.index(dep) + 1 if dep in sorted_plugins else -1
            max_index = max(idx, max_index)

        # Finally, insert after dependencies
        sorted_plugins.insert(max_index, plugin)


def get_plugin_deps(plugin_id: str) -> dict[str, Optional[str]]:
    result = {}
    try:
        parser = plugins_metadata_parser[plugin_id]
        plugin_deps = parser.get("general", "plugin_dependencies")
    except (configparser.NoOptionError, configparser.NoSectionError, KeyError):
        return result

    for dep in plugin_deps.split(","):
        if not dep:
            continue
        if "==" in dep:
            name, version_required = dep.split("==")
        else:
            name, version_required = dep, None
        result[name.strip()] = version_required.strip() if version_required else None

    return result


def pluginMetadata(packageName: str, fct: str) -> str:
    """fetch metadata from a plugin - use values from metadata.txt"""
    try:
        return plugins_metadata_parser[packageName].get("general", fct)
    except Exception:
        return "__error__"


def loadPlugin(packageName: str) -> bool:
    """load plugin's package"""

    try:
        __import__(packageName)
        return True
    except:
        pass  # continue...

    # snake in the grass, we know it's there
    sys.path_importer_cache.clear()

    # retry
    try:
        __import__(packageName)
        return True
    except:
        msg = QCoreApplication.translate("Python", "Couldn't load plugin '{0}'").format(
            packageName
        )
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
            level=Qgis.MessageLevel.Critical,
        )
        return False


def _startPlugin(packageName: str) -> bool:
    """initializes a plugin, but does not load GUI"""
    global plugins, active_plugins, iface, plugin_times

    if packageName in active_plugins:
        return False

    if packageName not in sys.modules:
        return False

    package = sys.modules[packageName]

    # Create the plugin's settings tree node and expose it as a module-level
    # attribute so the plugin can access it during classFactory and at module
    # import time. The node lives at "plugins/<packageName>" in the settings
    # tree and is automatically removed in unloadPlugin().
    try:
        package.SETTINGS_NODE = QgsSettingsTree.createPluginTreeNode(packageName)
    except Exception:
        QgsMessageLog.logMessage(
            f"Could not create settings tree node for plugin '{packageName}':\n{traceback.format_exc()}",
            "Plugins",
            Qgis.MessageLevel.Warning,
        )

    # create an instance of the plugin
    try:
        plugins[packageName] = package.classFactory(iface)
    except:
        # classFactory failed: drop the just-created settings node so we don't
        # leave a stale empty node behind.
        try:
            QgsSettingsTree.unregisterPluginTreeNode(packageName)
        except Exception:
            pass
        _unloadPluginModules(packageName)
        errMsg = QCoreApplication.translate(
            "Python", "Couldn't load plugin '{0}'"
        ).format(packageName)
        msg = QCoreApplication.translate(
            "Python", "{0} due to an error when calling its classFactory() method"
        ).format(errMsg)
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
            level=Qgis.MessageLevel.Critical,
        )
        return False
    return True


def _addToActivePlugins(packageName: str, duration: int):
    """Adds a plugin to the list of active plugins"""
    active_plugins.append(packageName)
    plugin_times[packageName] = f"{duration:02f}s"


def startPlugin(packageName: str) -> bool:
    """initialize the plugin"""
    global plugins, active_plugins, iface, plugin_times
    start = time.process_time()
    if not _startPlugin(packageName):
        return False

    # initGui
    try:
        plugins[packageName].initGui()
    except:
        del plugins[packageName]
        _unloadPluginModules(packageName)
        errMsg = QCoreApplication.translate(
            "Python", "Couldn't load plugin '{0}'"
        ).format(packageName)
        msg = QCoreApplication.translate(
            "Python", "{0} due to an error when calling its initGui() method"
        ).format(errMsg)
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
            level=Qgis.MessageLevel.Critical,
        )
        return False

    end = time.process_time()
    _addToActivePlugins(packageName, end - start)
    return True


def startProcessingPlugin(packageName: str) -> bool:
    """initialize only the Processing components of a plugin"""
    global plugins, active_plugins, iface, plugin_times
    start = time.process_time()
    if not _startPlugin(packageName):
        return False

    errMsg = QCoreApplication.translate("Python", "Couldn't load plugin '{0}'").format(
        packageName
    )
    if not hasattr(plugins[packageName], "initProcessing"):
        del plugins[packageName]
        _unloadPluginModules(packageName)
        msg = QCoreApplication.translate(
            "Python", "{0} - plugin has no initProcessing() method"
        ).format(errMsg)
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
            level=Qgis.MessageLevel.Critical,
        )
        return False

    # initProcessing
    try:
        plugins[packageName].initProcessing()
    except:
        del plugins[packageName]
        _unloadPluginModules(packageName)
        msg = QCoreApplication.translate(
            "Python", "{0} due to an error when calling its initProcessing() method"
        ).format(errMsg)
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
        )
        return False

    end = time.process_time()
    _addToActivePlugins(packageName, end - start)

    return True


def finalizeProcessingStartup() -> bool:
    """
    Finalizes the startup of the Processing plugin

    This should only be called after the startProcessingPlugin() method has been called
    for every installed and enabled plugin.
    """
    global plugins, active_plugins, iface, plugin_times
    if "processing" not in plugins:
        return False

    try:
        plugins["processing"].finalizeStartup()
    except:
        return False

    return True


def canUninstallPlugin(packageName: str) -> bool:
    """confirm that the plugin can be uninstalled"""
    global plugins, active_plugins

    if packageName not in plugins:
        return False
    if packageName not in active_plugins:
        return False

    try:
        metadata = plugins[packageName]
        if "canBeUninstalled" not in dir(metadata):
            return True
        return bool(metadata.canBeUninstalled())
    except:
        msg = "Error calling " + packageName + ".canBeUninstalled"
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
        )
        return True


def unloadPlugin(packageName: str) -> bool:
    """unload and delete plugin!"""
    global plugins, active_plugins

    if packageName not in plugins:
        return False
    if packageName not in active_plugins:
        return False

    try:
        plugins[packageName].unload()
        del plugins[packageName]
        active_plugins.remove(packageName)
        _unloadPluginModules(packageName)
        return True
    except Exception as e:
        msg = QCoreApplication.translate(
            "Python", "Error while unloading plugin {0}"
        ).format(packageName)
        showException(
            sys.exc_info()[0],
            sys.exc_info()[1],
            sys.exc_info()[2],
            msg,
            messagebar=True,
        )
        return False
    finally:
        # Always remove the plugin's settings tree node, even if unload() raised.
        try:
            QgsSettingsTree.unregisterPluginTreeNode(packageName)
        except Exception:
            QgsMessageLog.logMessage(
                f"Could not unregister settings tree node for plugin '{packageName}':\n{traceback.format_exc()}",
                "Plugins",
                Qgis.MessageLevel.Warning,
            )


def _unloadPluginModules(packageName: str):
    """unload plugin package with all its modules (files)"""
    global _plugin_modules

    if packageName not in _plugin_modules:
        return

    mods = _plugin_modules[packageName]

    for mod in mods:
        if mod not in sys.modules:
            continue

        # if it looks like a Qt resource file, try to do a cleanup
        # otherwise we might experience a segfault next time the plugin is loaded
        # because Qt will try to access invalid plugin resource data
        try:
            if hasattr(sys.modules[mod], "qCleanupResources"):
                sys.modules[mod].qCleanupResources()
        except:
            # Print stack trace for debug
            qDebug("qCleanupResources error:\n%s" % traceback.format_exc())

        # try removing path
        if hasattr(sys.modules[mod], "__path__"):
            for path in sys.modules[mod].__path__:
                try:
                    sys.path.remove(path)
                except ValueError:
                    # Discard if path is not there
                    pass

        # try to remove the module from python
        try:
            del sys.modules[mod]
        except:
            qDebug("Error when removing module:\n%s" % traceback.format_exc())
    # remove the plugin entry
    del _plugin_modules[packageName]


def isPluginLoaded(packageName: str) -> bool:
    """find out whether a plugin is active (i.e. has been started)"""
    global plugins, active_plugins

    if packageName not in plugins:
        return False
    return packageName in active_plugins


def reloadPlugin(packageName: str) -> bool:
    """unload and start again a plugin"""
    global active_plugins
    if packageName not in active_plugins:
        return False  # it's not active

    unloadPlugin(packageName)
    loadPlugin(packageName)
    started = startPlugin(packageName)
    return started


def showPluginHelp(packageName: str = None, filename: str = "index", section: str = ""):
    """Open help in the user's html browser. The help file should be named index-ll_CC.html or index-ll.html or index.html.

    :param str packageName: name of package folder, if None it's using the current file package. Defaults to None. Optional.
    :param str filename: name of file to open. It can be a path like 'doc/index' for example. Defaults to 'index'.
    :param str section: URL path to open. Defaults to empty string.
    """
    try:
        source = ""
        if packageName is None:
            import inspect

            source = inspect.currentframe().f_back.f_code.co_filename
        else:
            source = sys.modules[packageName].__file__
    except:
        return
    path = os.path.dirname(source)
    locale = str(QLocale().name())
    helpfile = os.path.join(path, filename + "-" + locale + ".html")
    if not os.path.exists(helpfile):
        helpfile = os.path.join(path, filename + "-" + locale.split("_")[0] + ".html")
    if not os.path.exists(helpfile):
        helpfile = os.path.join(path, filename + "-en.html")
    if not os.path.exists(helpfile):
        helpfile = os.path.join(path, filename + "-en_US.html")
    if not os.path.exists(helpfile):
        helpfile = os.path.join(path, filename + ".html")
    if os.path.exists(helpfile):
        url = "file://" + helpfile
        if section != "":
            url = url + "#" + section
        QDesktopServices.openUrl(QUrl.fromLocalFile(url))


def pluginDirectory(packageName: str) -> str:
    """return directory where the plugin resides. Plugin must be loaded already"""
    return os.path.dirname(sys.modules[packageName].__file__)


def reloadProjectMacros():
    # unload old macros
    unloadProjectMacros()

    from qgis.core import QgsProject

    code, ok = QgsProject.instance().readEntry("Macros", "/pythonCode")
    if not ok or not code or code == "":
        return

    # create a new empty python module
    import importlib

    mod = importlib.util.module_from_spec(
        importlib.machinery.ModuleSpec("proj_macros_mod", None)
    )

    # set the module code and store it sys.modules
    exec(str(code), mod.__dict__)
    sys.modules["proj_macros_mod"] = mod

    # load new macros
    openProjectMacro()


def unloadProjectMacros():
    if "proj_macros_mod" not in sys.modules:
        return
    # unload old macros
    closeProjectMacro()
    # destroy the reference to the module
    del sys.modules["proj_macros_mod"]


def openProjectMacro():
    if "proj_macros_mod" not in sys.modules:
        return
    mod = sys.modules["proj_macros_mod"]
    if hasattr(mod, "openProject"):
        mod.openProject()


def saveProjectMacro():
    if "proj_macros_mod" not in sys.modules:
        return
    mod = sys.modules["proj_macros_mod"]
    if hasattr(mod, "saveProject"):
        mod.saveProject()


def closeProjectMacro():
    if "proj_macros_mod" not in sys.modules:
        return
    mod = sys.modules["proj_macros_mod"]
    if hasattr(mod, "closeProject"):
        mod.closeProject()


#######################


def _list_project_expression_functions():
    """Get a list of expression functions stored in the current project"""
    import ast

    from qgis.core import QgsProject

    functions = []
    project_functions, ok = QgsProject.instance().readEntry(
        "ExpressionFunctions", "/pythonCode"
    )
    if ok and project_functions:
        code = ast.parse(project_functions)

        for e in code.body:
            if isinstance(e, ast.FunctionDef):
                for d in e.decorator_list:
                    if d.func.id == "qgsfunction":
                        functions.append(e.name)

    return functions


def clean_project_expression_functions():
    """
    Unload expression functions from current project
    and reload user expressions from profile folder
    to avoid any potential overwrite from the
    unloaded project functions
    """
    project_functions = _list_project_expression_functions()
    if project_functions:
        from qgis.core import QgsExpression

        for function in project_functions:
            QgsExpression.unregisterFunction(function)

        # Reload user expressions
        import expressions
        from qgis.core import QgsApplication

        userpythonhome = os.path.join(QgsApplication.qgisSettingsDirPath(), "python")
        expressionspath = os.path.join(userpythonhome, "expressions")
        expressions.reload(expressionspath)


#######################
# SERVER PLUGINS
#
# TODO: move into server_utils.py ?

# list of plugin paths. it gets filled in by the QGIS python library
server_plugin_paths = []

# dictionary of plugins
server_plugins = {}

# list of active (started) plugins
server_active_plugins = []


# initialize 'serverIface' object
serverIface = None


def initServerInterface(pointer):
    from qgis.PyQt.sip import wrapinstance
    from qgis.server import QgsServerInterface

    sys.excepthook = sys.__excepthook__
    global serverIface
    serverIface = wrapinstance(pointer, QgsServerInterface)


def startServerPlugin(packageName: str):
    """initialize the plugin"""
    global server_plugins, server_active_plugins, serverIface

    if packageName in server_active_plugins:
        return False
    if packageName not in sys.modules:
        return False

    package = sys.modules[packageName]

    errMsg = QCoreApplication.translate(
        "Python", "Couldn't load server plugin {0}"
    ).format(packageName)

    # create an instance of the plugin
    try:
        server_plugins[packageName] = package.serverClassFactory(serverIface)
    except:
        _unloadPluginModules(packageName)
        msg = QCoreApplication.translate(
            "Python", "{0} due to an error when calling its serverClassFactory() method"
        ).format(errMsg)
        showException(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2], msg)
        return False

    # add to active plugins
    server_active_plugins.append(packageName)
    return True


def spatialite_connect(*args, **kwargs):
    """returns a dbapi2.Connection to a SpatiaLite db
    using the "mod_spatialite" extension (python3)"""
    import re
    import sqlite3

    def fcnRegexp(pattern, string):
        result = re.search(pattern, string)
        return True if result else False

    con = sqlite3.dbapi2.connect(*args, **kwargs)
    con.enable_load_extension(True)
    cur = con.cursor()
    libs = [
        # SpatiaLite >= 4.2 and Sqlite >= 3.7.17, should work on all platforms
        ("mod_spatialite", "sqlite3_modspatialite_init"),
        # SpatiaLite >= 4.2 and Sqlite < 3.7.17 (Travis)
        ("mod_spatialite.so", "sqlite3_modspatialite_init"),
        # SpatiaLite < 4.2 (linux)
        ("libspatialite.so", "sqlite3_extension_init"),
    ]
    found = False
    for lib, entry_point in libs:
        try:
            cur.execute(f"select load_extension('{lib}', '{entry_point}')")
        except sqlite3.OperationalError:
            continue
        else:
            found = True
            break
    if not found:
        raise RuntimeError("Cannot find any suitable spatialite module")
    if any([".gpkg" in arg for arg in args]):
        try:
            cur.execute("SELECT EnableGpkgAmphibiousMode()")
        except (sqlite3.Error, sqlite3.DatabaseError, sqlite3.NotSupportedError):
            QgsMessageLog.logMessage(
                "warning:{}".format("Could not enable geopackage amphibious mode"),
                QCoreApplication.translate("Python", "Python warning"),
            )

    cur.close()
    con.enable_load_extension(False)
    con.create_function("regexp", 2, fcnRegexp)
    return con


class OverrideCursor:
    """
    Executes a code block with a different cursor set and makes sure the cursor
    is restored even if exceptions are raised or an intermediate ``return``
    statement is hit.

    Example:
    ```
    with OverrideCursor(Qt.WaitCursor):
        do_a_slow(operation)
    ```
    """

    def __init__(self, cursor):
        self.cursor = cursor

    def __enter__(self):
        QApplication.setOverrideCursor(self.cursor)

    def __exit__(self, exc_type, exc_val, exc_tb):
        QApplication.restoreOverrideCursor()
        return exc_type is None


#######################
# IMPORT wrapper

if os.name == "nt" and sys.version_info < (3, 8):
    import ctypes
    from ctypes import windll, wintypes

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    _hasAddDllDirectory = hasattr(kernel32, "AddDllDirectory")
    if _hasAddDllDirectory:
        _import_path = os.environ["PATH"]
        _import_paths = {}

        def _errcheck_zero(result, func, args):
            if not result:
                raise ctypes.WinError(ctypes.get_last_error())
            return args

        DLL_DIRECTORY_COOKIE = wintypes.LPVOID

        _AddDllDirectory = kernel32.AddDllDirectory
        _AddDllDirectory.errcheck = _errcheck_zero
        _AddDllDirectory.restype = DLL_DIRECTORY_COOKIE
        _AddDllDirectory.argtypes = (wintypes.LPCWSTR,)

        _RemoveDllDirectory = kernel32.RemoveDllDirectory
        _RemoveDllDirectory.errcheck = _errcheck_zero
        _RemoveDllDirectory.argtypes = (DLL_DIRECTORY_COOKIE,)

_uses_builtins = True
try:
    import builtins

    _builtin_import = builtins.__import__
except AttributeError:
    _uses_builtins = False
    import __builtin__

    _builtin_import = __builtin__.__import__

_plugin_modules = {}


def _import(name, globals={}, locals={}, fromlist=[], level=None):
    """
    Wrapper around builtin import that keeps track of loaded plugin modules and blocks
    certain unsafe imports
    """
    if level is None:
        level = 0

    qt_version = int(QT_VERSION_STR.split(".")[0])
    if qt_version == 6 and "PyQt5" in name:
        msg = (
            "PyQt5 classes cannot be imported in a QGIS build based on Qt6.\n"
            "Use {} or preferably the version independent {} import instead (where available).".format(
                name.replace("PyQt5", "PyQt6"), name.replace("PyQt5", "qgis.PyQt")
            )
        )
        raise ImportError(msg)

    if os.name == "nt" and sys.version_info < (3, 8):
        global _hasAddDllDirectory
        if _hasAddDllDirectory:
            global _import_path
            global _import_paths

            old_path = _import_path
            new_path = os.environ["PATH"]
            if old_path != new_path:
                global _AddDllDirectory
                global _RemoveDllDirectory

                for p in set(new_path.split(";")) - set(old_path.split(";")):
                    if p is not None and p not in _import_path and os.path.isdir(p):
                        _import_paths[p] = _AddDllDirectory(p)

                for p in set(old_path.split(";")) - set(new_path.split(";")):
                    if p in _import_paths:
                        _RemoveDllDirectory(_import_paths.pop(p))

                _import_path = new_path

    mod = _builtin_import(name, globals, locals, fromlist, level)

    if mod and getattr(mod, "__file__", None):
        module_name = mod.__name__ if fromlist else name
        package_name = module_name.split(".")[0]
        # check whether the module belongs to one of our plugins
        if package_name in available_plugins:
            if package_name not in _plugin_modules:
                _plugin_modules[package_name] = set()
            _plugin_modules[package_name].add(module_name)
            # check the fromlist for additional modules (from X import Y,Z)
            if fromlist:
                for fromitem in fromlist:
                    frmod = module_name + "." + fromitem
                    if frmod in sys.modules:
                        _plugin_modules[package_name].add(frmod)

    return mod


if not os.environ.get("QGIS_NO_OVERRIDE_IMPORT"):
    if _uses_builtins:
        builtins.__import__ = _import
    else:
        __builtin__.__import__ = _import


def processing_algorithm_from_script(filepath: str):
    """
    Tries to import a Python processing algorithm from given file, and returns an instance
    of the algorithm.

    Warning -- this ALSO execs the file as a script, so treat with caution!
    """
    import inspect
    import sys

    from qgis.processing import alg

    filename = filepath.replace("\\\\", "/")
    dirname = os.path.dirname(filename)

    # Append the directory of the file to the path and set __file__ to the filename
    added_script_dir_to_path = False
    if dirname not in sys.path:
        sys.path.append(dirname)
        added_script_dir_to_path = True

    alg_instance = None

    try:
        from qgis.core import (
            QgsApplication,
            QgsProcessingAlgorithm,
            QgsProcessingFeatureBasedAlgorithm,
        )

        _locals = {"__file__": filename}

        with open(filename.encode(sys.getfilesystemencoding())) as input_file:
            code_object = compile(input_file.read(), filename, "exec")
            exec(code_object, _locals)
        alg_instance = None
        try:
            alg_instance = alg.instances.pop().createInstance()
        except IndexError:
            for name, attr in _locals.items():
                if (
                    inspect.isclass(attr)
                    and issubclass(
                        attr,
                        (QgsProcessingAlgorithm, QgsProcessingFeatureBasedAlgorithm),
                    )
                    and attr.__name__
                    not in (
                        "QgsProcessingAlgorithm",
                        "QgsProcessingFeatureBasedAlgorithm",
                    )
                ):
                    alg_instance = attr()
                    break

    except ImportError:
        pass

    if alg_instance:
        script_provider = QgsApplication.processingRegistry().providerById("script")
        alg_instance.setProvider(script_provider)
        alg_instance.initAlgorithm()

    # gracefully clean up, restore previous system paths

    if added_script_dir_to_path:
        try:
            sys.path.remove(dirname)
        except ValueError:
            pass

    return alg_instance


def import_script_algorithm(filepath: str) -> Optional[str]:
    """
    Imports a script algorithm from given file to the processing script provider, and returns the
    ID of the imported algorithm.

    Warning -- this ALSO execs the file as a script, so treat with caution!
    """
    alg_instance = processing_algorithm_from_script(filepath)
    if alg_instance:
        from qgis.core import QgsApplication

        script_provider = QgsApplication.processingRegistry().providerById("script")
        script_provider.add_algorithm_class(type(alg_instance))
        return alg_instance.id()

    return None


def run_script_from_file(filepath: str):
    """
    Runs a Python script from a given file. Supports loading processing scripts.
    :param filepath: The .py file to load.
    """
    try:
        from qgis.processing import execAlgorithmDialog
    except ImportError:
        return

    # Despite the misleading name, processing_algorithm_from_script will ALWAYS
    # execute the script from the file!
    alg_instance = processing_algorithm_from_script(filepath)
    if alg_instance:
        execAlgorithmDialog(alg_instance)



"""
***************************************************************************
    pct2rgb.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsRasterFileWriter,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils
from processing.tools.system import isWindows

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class pct2rgb(GdalAlgorithm):
    INPUT = "INPUT"
    BAND = "BAND"
    RGBA = "RGBA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBand(
                self.BAND,
                self.tr("Band number"),
                1,
                parentLayerParameterName=self.INPUT,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.RGBA, self.tr("Generate a RGBA file"), defaultValue=False
            )
        )
        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("PCT to RGB"))
        )

    def name(self):
        return "pcttorgb"

    def displayName(self):
        return self.tr("PCT to RGB")

    def group(self):
        return self.tr("Raster conversion")

    def groupId(self):
        return "rasterconversion"

    def icon(self):
        return QIcon(
            os.path.join(pluginPath, "images", "gdaltools", "8-to-24-bits.png")
        )

    def commandName(self):
        return "pct2rgb"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        input_details = GdalUtils.gdal_connection_details_from_layer(inLayer)

        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        output_format = self.outputFormat(parameters, self.OUTPUT, context)
        if not output_format:
            raise QgsProcessingException(self.tr("Output format is invalid"))

        arguments = [
            input_details.connection_string,
            out,
            "-of",
            output_format,
            "-b",
            str(self.parameterAsInt(parameters, self.BAND, context)),
        ]

        if self.parameterAsBoolean(parameters, self.RGBA, context):
            arguments.append("-rgba")

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        return [
            self.commandName() + (".bat" if isWindows() else ".py"),
            GdalUtils.escapeAndJoin(arguments),
        ]



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

The content of this file is based on
- PG_Manager by Martin Dobias (GPLv2 license)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import functools

from qgis.core import Qgis, QgsApplication, QgsMapLayerType, QgsSettings
from qgis.gui import QgsMessageBar
from qgis.PyQt.QtCore import QByteArray, QSize, Qt
from qgis.PyQt.QtGui import QIcon, QKeySequence
from qgis.PyQt.QtWidgets import (
    QAction,
    QApplication,
    QDockWidget,
    QGridLayout,
    QMainWindow,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QSpacerItem,
    QStatusBar,
    QTabBar,
    QTabWidget,
    QToolBar,
)
from qgis.utils import OverrideCursor

from .db_plugins.plugin import BaseError
from .db_tree import DBTree
from .dlg_db_error import DlgDbError
from .gui_utils import GuiUtils
from .info_viewer import InfoViewer
from .layer_preview import LayerPreview
from .table_viewer import TableViewer


class DBManager(QMainWindow):
    def __init__(self, iface, parent=None):
        QMainWindow.__init__(self, parent)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setupUi()
        self.iface = iface

        # restore the window state
        settings = QgsSettings()
        self.restoreGeometry(
            settings.value(
                "/DB_Manager/mainWindow/geometry", QByteArray(), type=QByteArray
            )
        )
        self.restoreState(
            settings.value(
                "/DB_Manager/mainWindow/windowState", QByteArray(), type=QByteArray
            )
        )

        self.toolBar.setIconSize(self.iface.iconSize())
        self.toolBarOrientation()
        self.toolBar.orientationChanged.connect(self.toolBarOrientation)
        self.tabs.currentChanged.connect(self.tabChanged)
        self.tree.selectedItemChanged.connect(self.itemChanged)
        self.tree.model().dataChanged.connect(self.iface.reloadConnections)
        self.itemChanged(None)

    def closeEvent(self, e):
        self.unregisterAllActions()
        # clear preview, this will delete the layer in preview tab
        self.preview.loadPreview(None)

        # save the window state
        settings = QgsSettings()
        settings.setValue("/DB_Manager/mainWindow/windowState", self.saveState())
        settings.setValue("/DB_Manager/mainWindow/geometry", self.saveGeometry())

        QMainWindow.closeEvent(self, e)

    def refreshItem(self, item=None):
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                if item is None:
                    item = self.tree.currentItem()
                self.tree.refreshItem(item)  # refresh item children in the db tree
            except BaseError as e:
                DlgDbError.showError(e, self)

    def itemChanged(self, item):
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                self.reloadButtons()
                # Force-reload information on the layer
                self.info.setDirty()
                # clear preview, this will delete the layer in preview tab
                self.preview.loadPreview(None)
                self.refreshTabs()
            except BaseError as e:
                DlgDbError.showError(e, self)

    def reloadButtons(self):
        db = self.tree.currentDatabase()
        if not hasattr(self, "_lastDb"):
            self._lastDb = db

        elif db == self._lastDb:
            return

        # remove old actions
        if self._lastDb is not None:
            self.unregisterAllActions()

        # add actions of the selected database
        self._lastDb = db
        if self._lastDb is not None:
            self._lastDb.registerAllActions(self)

    def tabChanged(self, index):
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                self.refreshTabs()
            except BaseError as e:
                DlgDbError.showError(e, self)

    def refreshTabs(self):
        index = self.tabs.currentIndex()
        item = self.tree.currentItem()
        table = self.tree.currentTable()

        # enable/disable tabs
        self.tabs.setTabEnabled(self.tabs.indexOf(self.table), table is not None)
        self.tabs.setTabEnabled(
            self.tabs.indexOf(self.preview),
            table is not None
            and table.type in [table.VectorType, table.RasterType]
            and table.geomColumn is not None,
        )
        # show the info tab if the current tab is disabled
        if not self.tabs.isTabEnabled(index):
            self.tabs.setCurrentWidget(self.info)

        current_tab = self.tabs.currentWidget()
        if current_tab == self.info:
            self.info.showInfo(item)
        elif current_tab == self.table:
            self.table.loadData(item)
        elif current_tab == self.preview:
            self.preview.loadPreview(item)

    def refreshActionSlot(self):
        self.info.setDirty()
        self.table.setDirty()
        self.preview.setDirty()
        self.refreshItem()

    def importActionSlot(self):
        db = self.tree.currentDatabase()
        if db is None:
            self.infoBar.pushMessage(
                self.tr("No database selected or you are not connected to it."),
                Qgis.MessageLevel.Info,
                self.iface.messageTimeout(),
            )
            return

        outUri = db.uri()
        schema = self.tree.currentSchema()
        if schema:
            outUri.setDataSource(schema.name, "", "", "")

        from .dlg_import_vector import DlgImportVector

        dlg = DlgImportVector(None, db, outUri, self)
        dlg.exec()

    def exportActionSlot(self):
        table = self.tree.currentTable()
        if table is None:
            self.infoBar.pushMessage(
                self.tr("Select the table you want export to file."),
                Qgis.MessageLevel.Info,
                self.iface.messageTimeout(),
            )
            return

        inLayer = table.toMapLayer()
        if inLayer.type() != QgsMapLayerType.VectorLayer:
            self.infoBar.pushMessage(
                self.tr("Select a vector or a tabular layer you want export."),
                Qgis.MessageLevel.Warning,
                self.iface.messageTimeout(),
            )
            return

        from .dlg_export_vector import DlgExportVector

        dlg = DlgExportVector(inLayer, table.database(), self)
        dlg.exec()

        inLayer.deleteLater()

    def runSqlWindow(self):
        db = self.tree.currentDatabase()
        if db is None:
            self.infoBar.pushMessage(
                self.tr("No database selected or you are not connected to it."),
                Qgis.MessageLevel.Info,
                self.iface.messageTimeout(),
            )
            # force displaying of the message, it appears on the first tab (i.e. Info)
            self.tabs.setCurrentIndex(0)
            return

        from .dlg_sql_window import DlgSqlWindow

        query = DlgSqlWindow(self.iface, db, self)
        dbname = db.connection().connectionName()
        tabname = self.tr("Query ({0})").format(dbname)
        index = self.tabs.addTab(query, tabname)
        self.tabs.setTabIcon(index, db.connection().icon())
        self.tabs.setCurrentIndex(index)
        query.nameChanged.connect(
            functools.partial(self.update_query_tab_name, index, dbname)
        )

    def update_query_tab_name(self, index, dbname, queryname):
        if not queryname:
            queryname = self.tr("Query")
        tabname = f"{queryname} ({dbname})"
        self.tabs.setTabText(index, tabname)

    def showSystemTables(self):
        self.tree.showSystemTables(self.actionShowSystemTables.isChecked())

    def registerAction(self, action, menuName, callback=None):
        """register an action to the manager's main menu"""
        if not hasattr(self, "_registeredDbActions"):
            self._registeredDbActions = {}

        if callback is not None:

            def invoke_callback(x):
                return self.invokeCallback(callback)

        if menuName is None or menuName == "":
            self.addAction(action)

            if menuName not in self._registeredDbActions:
                self._registeredDbActions[menuName] = list()
            self._registeredDbActions[menuName].append(action)

            if callback is not None:
                action.triggered.connect(invoke_callback)
            return True

        # search for the menu
        actionMenu = None
        helpMenuAction = None
        for a in self.menuBar.actions():
            if not a.menu() or a.menu().title() != menuName:
                continue
            if a.menu() != self.menuHelp:
                helpMenuAction = a

            actionMenu = a
            break

        # not found, add a new menu before the help menu
        if actionMenu is None:
            menu = QMenu(menuName, self)
            if helpMenuAction is not None:
                actionMenu = self.menuBar.insertMenu(helpMenuAction, menu)
            else:
                actionMenu = self.menuBar.addMenu(menu)

        menu = actionMenu.menu()
        menuActions = menu.actions()

        # get the placeholder's position to insert before it
        pos = 0
        for pos in range(len(menuActions)):
            if menuActions[pos].isSeparator() and menuActions[
                pos
            ].objectName().endswith("_placeholder"):
                menuActions[pos].setVisible(True)
                break

        if pos < len(menuActions):
            before = menuActions[pos]
            menu.insertAction(before, action)
        else:
            menu.addAction(action)

        actionMenu.setVisible(True)  # show the menu

        if menuName not in self._registeredDbActions:
            self._registeredDbActions[menuName] = list()
        self._registeredDbActions[menuName].append(action)

        if callback is not None:
            action.triggered.connect(invoke_callback)

        return True

    def invokeCallback(self, callback, *params):
        """Call a method passing the selected item in the database tree,
        the sender (usually a QAction), the plugin mainWindow and
        optionally additional parameters.

        This method takes care to override and restore the cursor,
        but also catches exceptions and displays the error dialog.
        """
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                callback(self.tree.currentItem(), self.sender(), self, *params)
            except BaseError as e:
                # catch database errors and display the error dialog
                DlgDbError.showError(e, self)

    def unregisterAction(self, action, menuName):
        if not hasattr(self, "_registeredDbActions"):
            return

        if menuName is None or menuName == "":
            self.removeAction(action)

            if menuName in self._registeredDbActions:
                if self._registeredDbActions[menuName].count(action) > 0:
                    self._registeredDbActions[menuName].remove(action)

            action.deleteLater()
            return True

        for a in self.menuBar.actions():
            if not a.menu() or a.menu().title() != menuName:
                continue

            menu = a.menu()
            menuActions = menu.actions()

            menu.removeAction(action)
            if menu.isEmpty():  # hide the menu
                a.setVisible(False)

            if menuName in self._registeredDbActions:
                if self._registeredDbActions[menuName].count(action) > 0:
                    self._registeredDbActions[menuName].remove(action)

                # hide the placeholder if there're no other registered actions
                if len(self._registeredDbActions[menuName]) <= 0:
                    for i in range(len(menuActions)):
                        if menuActions[i].isSeparator() and menuActions[
                            i
                        ].objectName().endswith("_placeholder"):
                            menuActions[i].setVisible(False)
                            break

            action.deleteLater()
            return True

        return False

    def unregisterAllActions(self):
        if not hasattr(self, "_registeredDbActions"):
            return

        for menuName in self._registeredDbActions:
            for action in list(self._registeredDbActions[menuName]):
                self.unregisterAction(action, menuName)
        del self._registeredDbActions

    def close_tab(self, index):
        widget = self.tabs.widget(index)
        if widget not in [self.info, self.table, self.preview]:
            if hasattr(widget, "close"):
                if widget.close():
                    self.tabs.removeTab(index)
                    widget.deleteLater()
            else:
                self.tabs.removeTab(index)
                widget.deleteLater()

    def toolBarOrientation(self):
        button_style = Qt.ToolButtonStyle.ToolButtonIconOnly
        if self.toolBar.orientation() == Qt.Orientation.Horizontal:
            button_style = Qt.ToolButtonStyle.ToolButtonTextBesideIcon

        widget = self.toolBar.widgetForAction(self.actionImport)
        widget.setToolButtonStyle(button_style)
        widget = self.toolBar.widgetForAction(self.actionExport)
        widget.setToolButtonStyle(button_style)

    def setupUi(self):
        self.setWindowTitle(self.tr("DB Manager"))
        self.setWindowIcon(GuiUtils.get_icon("dbmanager"))
        self.resize(QSize(700, 500).expandedTo(self.minimumSizeHint()))

        # create central tab widget and add the first 3 tabs: info, table and preview
        self.tabs = QTabWidget()
        self.info = InfoViewer(self)
        self.tabs.addTab(self.info, self.tr("Info"))
        self.table = TableViewer(self)
        self.tabs.addTab(self.table, self.tr("Table"))
        self.preview = LayerPreview(self)
        self.tabs.addTab(self.preview, self.tr("Preview"))
        self.setCentralWidget(self.tabs)

        # display close button for all tabs but the first 3 ones, i.e.
        # HACK: just hide the close button where not needed (GS)
        self.tabs.setTabsClosable(True)
        self.tabs.tabCloseRequested.connect(self.close_tab)
        tabbar = self.tabs.tabBar()
        for i in range(3):
            btn = (
                tabbar.tabButton(i, QTabBar.ButtonPosition.RightSide)
                if tabbar.tabButton(i, QTabBar.ButtonPosition.RightSide)
                else tabbar.tabButton(i, QTabBar.ButtonPosition.LeftSide)
            )
            btn.resize(0, 0)
            btn.hide()

        # Creates layout for message bar
        self.layout = QGridLayout(self.info)
        self.layout.setContentsMargins(0, 0, 0, 0)
        spacerItem = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        self.layout.addItem(spacerItem, 1, 0, 1, 1)
        # init messageBar instance
        self.infoBar = QgsMessageBar(self.info)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.infoBar.setSizePolicy(sizePolicy)
        self.layout.addWidget(self.infoBar, 0, 0, 1, 1)

        # create database tree
        self.dock = QDockWidget(self.tr("Providers"), self)
        self.dock.setObjectName("DB_Manager_DBView")
        self.dock.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.tree = DBTree(self)
        self.dock.setWidget(self.tree)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.dock)

        # create status bar
        self.statusBar = QStatusBar(self)
        self.setStatusBar(self.statusBar)

        # create menus
        self.menuBar = QMenuBar(self)
        self.menuDb = QMenu(self.tr("&Database"), self)
        self.menuBar.addMenu(self.menuDb)
        self.menuSchema = QMenu(self.tr("&Schema"), self)
        actionMenuSchema = self.menuBar.addMenu(self.menuSchema)
        self.menuTable = QMenu(self.tr("&Table"), self)
        actionMenuTable = self.menuBar.addMenu(self.menuTable)
        self.menuHelp = None  # QMenu(self.tr("&Help"), self)
        # actionMenuHelp = self.menuBar.addMenu(self.menuHelp)

        self.setMenuBar(self.menuBar)

        # create toolbar
        self.toolBar = QToolBar(self.tr("Default"), self)
        self.toolBar.setObjectName("DB_Manager_ToolBar")
        self.addToolBar(self.toolBar)

        # create menus' actions

        # menu DATABASE
        sep = self.menuDb.addSeparator()
        sep.setObjectName("DB_Manager_DbMenu_placeholder")
        sep.setVisible(False)

        self.actionRefresh = QAction(
            QgsApplication.getThemeIcon("/mActionRefresh.svg"),
            self.tr("&Refresh"),
            self.menuDb,
        )
        self.actionRefresh.triggered.connect(self.refreshActionSlot)
        self.actionRefresh.setShortcut(QKeySequence("F5"))
        self.menuDb.addAction(self.actionRefresh)

        self.actionSqlWindow = QAction(
            GuiUtils.get_icon("mActionSQLWindow"), self.tr("&SQL Window"), self.menuDb
        )
        self.actionSqlWindow.triggered.connect(self.runSqlWindow)
        self.actionSqlWindow.setShortcut(QKeySequence("F2"))
        self.menuDb.addAction(self.actionSqlWindow)

        self.menuDb.addSeparator()

        self.actionClose = QAction(QIcon(), self.tr("&Exit"), self.menuDb)
        self.actionClose.triggered.connect(self.close)
        self.actionClose.setShortcut(QKeySequence("CTRL+Q"))
        self.menuDb.addAction(self.actionClose)

        # menu SCHEMA
        sep = self.menuSchema.addSeparator()
        sep.setObjectName("DB_Manager_SchemaMenu_placeholder")
        sep.setVisible(False)

        actionMenuSchema.setVisible(False)

        # menu TABLE
        sep = self.menuTable.addSeparator()
        sep.setObjectName("DB_Manager_TableMenu_placeholder")
        sep.setVisible(False)

        self.actionImport = self.menuTable.addAction(
            GuiUtils.get_icon("mActionDBImport"),
            QApplication.translate("DBManager", "&Import Layer/File…"),
            self.importActionSlot,
        )
        self.actionExport = self.menuTable.addAction(
            GuiUtils.get_icon("mActionDBExport"),
            QApplication.translate("DBManager", "&Export to File…"),
            self.exportActionSlot,
        )
        self.menuTable.addSeparator()
        # self.actionShowSystemTables = self.menuTable.addAction(self.tr("Show system tables/views"), self.showSystemTables)
        # self.actionShowSystemTables.setCheckable(True)
        # self.actionShowSystemTables.setChecked(True)
        actionMenuTable.setVisible(False)

        # add actions to the toolbar
        self.toolBar.addAction(self.actionRefresh)
        self.toolBar.addAction(self.actionSqlWindow)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionImport)
        self.toolBar.addAction(self.actionExport)



"""
***************************************************************************
    ModelerParametersDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import webbrowser
from typing import Optional

from qgis.core import (
    Qgis,
    QgsProcessingAlgorithm,
    QgsProcessingModelAlgorithm,
    QgsProcessingModelChildAlgorithm,
    QgsProcessingModelChildParameterSource,
    QgsProcessingModelOutput,
    QgsProcessingOutputDefinition,
    QgsProcessingParameterDefinition,
    QgsProject,
)
from qgis.gui import (
    QgsCollapsibleGroupBox,
    QgsColorButton,
    QgsFilterLineEdit,
    QgsGui,
    QgsHelp,
    QgsMessageBar,
    QgsModelChildDependenciesWidget,
    QgsPanelWidget,
    QgsPanelWidgetStack,
    QgsProcessingContextGenerator,
    QgsProcessingModelConfigWidget,
    QgsProcessingModelerParameterWidget,
    QgsProcessingParameterWidgetContext,
    QgsScrollArea,
)
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QColor
from qgis.PyQt.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from qgis.utils import iface

from processing.core.exceptions import InvalidParameterValue
from processing.gui.wrappers import (
    WidgetWrapper,
    WidgetWrapperFactory,
)
from processing.tools.dataobjects import createContext


class ModelerParametersDialog(QDialog):
    def __init__(
        self,
        alg: QgsProcessingAlgorithm,
        model: QgsProcessingModelAlgorithm,
        algName: Optional[str] = None,
        configuration: Optional[dict[str, object]] = None,
    ):
        super().__init__()
        self.setObjectName("ModelerParametersDialog")

        if iface is not None:
            self.setStyleSheet(iface.mainWindow().styleSheet())

        # dammit this is SUCH as mess... stupid stable API
        self._alg = alg  # The algorithm to define in this dialog. It is an instance of QgsProcessingAlgorithm
        self.model = model  # The model this algorithm is going to be added to. It is an instance of QgsProcessingModelAlgorithm
        self.childId = algName  # The name of the algorithm in the model, in case we are editing it and not defining it for the first time
        self.configuration = configuration
        self.context = createContext()

        self.setWindowTitle(
            " - ".join([self._alg.group(), self._alg.displayName()])
            if self._alg.group()
            else self._alg.displayName()
        )

        self.widget = ModelerParametersWidget(
            alg, model, algName, configuration, context=self.context, dialog=self
        )
        QgsGui.enableAutoGeometryRestore(self)

        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.StandardButton.Cancel
            | QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Help
        )

        self.buttonBox.accepted.connect(self.okPressed)
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.helpRequested.connect(self.openHelp)

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self.widget, 1)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def algorithm(self):
        return self._alg

    def setComments(self, text):
        self.widget.setComments(text)

    def comments(self):
        return self.widget.comments()

    def setCommentColor(self, color):
        self.widget.setCommentColor(color)

    def commentColor(self):
        return self.widget.commentColor()

    def switchToCommentTab(self):
        self.widget.switchToCommentTab()

    def getAvailableValuesOfType(self, paramType, outTypes=[], dataTypes=[]):
        # upgrade paramType to list
        if paramType is None:
            paramType = []
        elif not isinstance(paramType, (tuple, list)):
            paramType = [paramType]
        if outTypes is None:
            outTypes = []
        elif not isinstance(outTypes, (tuple, list)):
            outTypes = [outTypes]

        return self.model.availableSourcesForChild(
            self.childId,
            [
                p.typeName()
                for p in paramType
                if issubclass(p, QgsProcessingParameterDefinition)
            ],
            [
                o.typeName()
                for o in outTypes
                if issubclass(o, QgsProcessingOutputDefinition)
            ],
            dataTypes,
        )

    def resolveValueDescription(self, value):
        if isinstance(value, QgsProcessingModelChildParameterSource):
            if value.source() == Qgis.ProcessingModelChildParameterSource.StaticValue:
                return value.staticValue()
            elif (
                value.source()
                == Qgis.ProcessingModelChildParameterSource.ModelParameter
            ):
                return self.model.parameterDefinition(
                    value.parameterName()
                ).description()
            elif value.source() == Qgis.ProcessingModelChildParameterSource.ChildOutput:
                alg = self.model.childAlgorithm(value.outputChildId())

                output_name = (
                    alg.algorithm().outputDefinition(value.outputName()).description()
                )
                # see if this output has been named by the model designer -- if so, we use that friendly name
                for name, output in alg.modelOutputs().items():
                    if output.childOutputName() == value.outputName():
                        output_name = name
                        break

                return self.tr("'{0}' from algorithm '{1}'").format(
                    output_name, alg.description()
                )

        return value

    def setPreviousValues(self):
        self.widget.setPreviousValues()

    def createAlgorithm(self):
        return self.widget.createAlgorithm()

    def okPressed(self):
        if self.createAlgorithm() is not None:
            self.accept()

    def openHelp(self):
        algHelp = self.widget.algorithm().helpUrl()
        if not algHelp:
            algHelp = QgsHelp.helpUrl(
                "processing_algs/{}/{}.html#{}".format(
                    self.widget.algorithm().provider().helpId(),
                    self.algorithm().groupId(),
                    f"{self.algorithm().provider().helpId()}{self.algorithm().name().replace('_', '-')}",
                )
            ).toString()

        if algHelp not in [None, ""]:
            webbrowser.open(algHelp)


class ModelerParametersPanelWidget(QgsPanelWidget):
    def __init__(
        self, alg, model, algName=None, configuration=None, dialog=None, context=None
    ):
        super().__init__()
        self._alg = alg  # The algorithm to define in this dialog. It is an instance of QgsProcessingAlgorithm
        self.model = model  # The model this algorithm is going to be added to. It is an instance of QgsProcessingModelAlgorithm
        self.childId = algName  # The name of the algorithm in the model, in case we are editing it and not defining it for the first time
        self.configuration = configuration
        self.context = context
        self.dialog = dialog
        self.previous_output_definitions = {}
        self.block_changes_signal = 0

        class ContextGenerator(QgsProcessingContextGenerator):
            def __init__(self, context):
                super().__init__()
                self.processing_context = context

            def processingContext(self):
                return self.processing_context

        self.context_generator = ContextGenerator(self.context)

        self.setupUi()
        self.params = None

    def algorithm(self):
        return self._alg

    def setupUi(self):
        self.showAdvanced = False
        self.wrappers = {}
        self.algorithmItem = None

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)

        self.verticalLayout = QVBoxLayout()

        self.bar = QgsMessageBar()
        self.bar.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.verticalLayout.addWidget(self.bar)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        descriptionLabel = QLabel(self.tr("Description"))
        self.descriptionBox = QLineEdit()
        self.descriptionBox.setText(self._alg.displayName())
        hLayout.addWidget(descriptionLabel)
        hLayout.addWidget(self.descriptionBox)
        self.descriptionBox.textChanged.connect(self.emit_changed_signal)

        self.verticalLayout.addLayout(hLayout)
        line = QFrame()
        line.setFrameShape(QFrame.Shape.HLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)
        self.verticalLayout.addWidget(line)

        widget_context = QgsProcessingParameterWidgetContext()
        widget_context.setProject(QgsProject.instance())
        if iface is not None:
            widget_context.setMapCanvas(iface.mapCanvas())
            widget_context.setActiveLayer(iface.activeLayer())

        widget_context.setModel(self.model)
        widget_context.setModelChildAlgorithmId(self.childId)

        self.algorithmItem = (
            QgsGui.instance()
            .processingGuiRegistry()
            .algorithmConfigurationWidget(self._alg)
        )
        if self.algorithmItem:
            self.algorithmItem.setWidgetContext(widget_context)
            self.algorithmItem.registerProcessingContextGenerator(
                self.context_generator
            )
            if self.configuration:
                self.algorithmItem.setConfiguration(self.configuration)
            self.verticalLayout.addWidget(self.algorithmItem)

        self.grpAdvanced = QgsCollapsibleGroupBox(self.tr("Advanced Parameters"))
        self.grpAdvancedVLayout = QVBoxLayout()
        self.grpAdvanced.setLayout(self.grpAdvancedVLayout)
        self.grpAdvanced.hide()

        self.verticalLayout.addWidget(self.grpAdvanced)

        for param in self._alg.parameterDefinitions():
            if param.flags() & QgsProcessingParameterDefinition.Flag.FlagAdvanced:
                self.grpAdvanced.show()
                break
        for param in self._alg.parameterDefinitions():
            if (
                param.isDestination()
                or param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden
            ):
                continue

            wrapper = WidgetWrapperFactory.create_wrapper(param, self.dialog)
            self.wrappers[param.name()] = wrapper

            wrapper.setWidgetContext(widget_context)
            wrapper.registerProcessingContextGenerator(self.context_generator)
            if issubclass(wrapper.__class__, QgsProcessingModelerParameterWidget):
                widget = wrapper
                wrapper.changed.connect(self.emit_changed_signal)
            else:
                widget = wrapper.widget
            if widget is not None:
                if issubclass(wrapper.__class__, QgsProcessingModelerParameterWidget):
                    label = wrapper.createLabel()
                else:
                    tooltip = param.description()
                    widget.setToolTip(tooltip)
                    label = wrapper.label

                if param.flags() & QgsProcessingParameterDefinition.Flag.FlagAdvanced:
                    self.grpAdvancedVLayout.addWidget(label)
                    self.grpAdvancedVLayout.addWidget(widget)
                else:
                    # Regular parameters
                    self.verticalLayout.insertWidget(
                        self.verticalLayout.count() - 1, label
                    )
                    self.verticalLayout.insertWidget(
                        self.verticalLayout.count() - 1, widget
                    )

        for output in self._alg.destinationParameterDefinitions():
            if output.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                continue

            widget = QgsGui.processingGuiRegistry().createModelerParameterWidget(
                self.model, self.childId, output, self.context
            )
            widget.setDialog(self.dialog)
            widget.setWidgetContext(widget_context)
            widget.registerProcessingContextGenerator(self.context_generator)
            if isinstance(widget, QgsProcessingModelerParameterWidget):
                widget.changed.connect(self.emit_changed_signal)

            self.wrappers[output.name()] = widget

            label = widget.createLabel()
            if label is not None:
                self.verticalLayout.addWidget(label)

            self.verticalLayout.addWidget(widget)

        label = QLabel(" ")
        self.verticalLayout.addWidget(label)
        label = QLabel(self.tr("Dependencies"))
        self.dependencies_panel = QgsModelChildDependenciesWidget(
            self, self.model, self.childId
        )
        self.verticalLayout.addWidget(label)
        self.verticalLayout.addWidget(self.dependencies_panel)
        self.verticalLayout.addStretch(1000)

        self.setPreviousValues()
        self.verticalLayout2 = QVBoxLayout()
        self.verticalLayout2.setSpacing(2)
        self.verticalLayout2.setMargin(0)

        self.paramPanel = QWidget()
        self.paramPanel.setLayout(self.verticalLayout)
        self.scrollArea = QgsScrollArea()
        self.scrollArea.setWidget(self.paramPanel)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setFrameStyle(QFrame.Shape.NoFrame)

        self.verticalLayout2.addWidget(self.scrollArea)

        w = QWidget()
        w.setLayout(self.verticalLayout2)
        self.mainLayout.addWidget(w)
        self.setLayout(self.mainLayout)

    def emit_changed_signal(self):
        if not self.block_changes_signal:
            self.widgetChanged.emit()

    def showAdvancedParametersClicked(self):
        self.showAdvanced = not self.showAdvanced
        if self.showAdvanced:
            self.advancedButton.setText(self.tr("Hide advanced parameters"))
        else:
            self.advancedButton.setText(self.tr("Show advanced parameters"))
        for param in self._alg.parameterDefinitions():
            if param.flags() & QgsProcessingParameterDefinition.Flag.FlagAdvanced:
                wrapper = self.wrappers[param.name()]
                if issubclass(wrapper.__class__, QgsProcessingModelerParameterWidget):
                    wrapper.setVisible(self.showAdvanced)
                else:
                    wrapper.widget.setVisible(self.showAdvanced)

                self.widget_labels[param.name()].setVisible(self.showAdvanced)

    def setPreviousValues(self):
        self.block_changes_signal += 1
        if self.childId is not None:
            alg = self.model.childAlgorithm(self.childId)

            self.descriptionBox.setText(alg.description())
            for param in alg.algorithm().parameterDefinitions():
                if (
                    param.isDestination()
                    or param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden
                ):
                    continue
                value = None
                if param.name() in alg.parameterSources():
                    value = alg.parameterSources()[param.name()]
                    if isinstance(value, list) and len(value) == 1:
                        value = value[0]
                    elif isinstance(value, list) and len(value) == 0:
                        value = None

                wrapper = self.wrappers[param.name()]
                if issubclass(wrapper.__class__, QgsProcessingModelerParameterWidget):
                    if value is None:
                        value = QgsProcessingModelChildParameterSource.fromStaticValue(
                            param.defaultValue()
                        )

                    wrapper.setWidgetValue(value)
                else:
                    if value is None:
                        value = param.defaultValue()

                    if (
                        isinstance(value, QgsProcessingModelChildParameterSource)
                        and value.source()
                        == Qgis.ProcessingModelChildParameterSource.StaticValue
                    ):
                        value = value.staticValue()
                    wrapper.setValue(value)

            for output in self.algorithm().destinationParameterDefinitions():
                if output.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                    continue

                model_output_name = None
                for name, out in alg.modelOutputs().items():
                    if (
                        out.childId() == self.childId
                        and out.childOutputName() == output.name()
                    ):
                        # this destination parameter is linked to a model output
                        model_output_name = out.name()
                        self.previous_output_definitions[output.name()] = out
                        break

                value = None
                if (
                    model_output_name is None
                    and output.name() in alg.parameterSources()
                ):
                    value = alg.parameterSources()[output.name()]
                    if isinstance(value, list) and len(value) == 1:
                        value = value[0]
                    elif isinstance(value, list) and len(value) == 0:
                        value = None

                wrapper = self.wrappers[output.name()]

                if model_output_name is not None:
                    wrapper.setToModelOutput(model_output_name)
                elif value is not None or output.defaultValue() is not None:
                    if value is None:
                        value = QgsProcessingModelChildParameterSource.fromStaticValue(
                            output.defaultValue()
                        )

                    wrapper.setWidgetValue(value)

            self.dependencies_panel.setValue(alg.dependencies())
        self.block_changes_signal -= 1

    def createAlgorithm(self):
        alg = QgsProcessingModelChildAlgorithm(self._alg.id())
        if not self.childId:
            alg.generateChildId(self.model)
        else:
            alg.setChildId(self.childId)
        alg.setDescription(self.descriptionBox.text())
        if self.algorithmItem:
            alg.setConfiguration(self.algorithmItem.configuration())
            self._alg = alg.algorithm().create(self.algorithmItem.configuration())
        for param in self._alg.parameterDefinitions():
            if (
                param.isDestination()
                or param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden
            ):
                continue
            try:
                wrapper = self.wrappers[param.name()]
                if issubclass(wrapper.__class__, WidgetWrapper):
                    val = wrapper.value()
                elif issubclass(wrapper.__class__, QgsProcessingModelerParameterWidget):
                    val = wrapper.value()
                else:
                    val = wrapper.parameterValue()
            except InvalidParameterValue:
                val = None

            if isinstance(val, QgsProcessingModelChildParameterSource):
                val = [val]
            elif not (
                isinstance(val, list)
                and all(
                    [
                        isinstance(subval, QgsProcessingModelChildParameterSource)
                        for subval in val
                    ]
                )
            ):
                val = [QgsProcessingModelChildParameterSource.fromStaticValue(val)]

            valid = True
            for subval in val:
                if (
                    isinstance(subval, QgsProcessingModelChildParameterSource)
                    and subval.source()
                    == Qgis.ProcessingModelChildParameterSource.StaticValue
                    and not param.checkValueIsAcceptable(subval.staticValue())
                ) or (
                    subval is None
                    and not param.flags()
                    & QgsProcessingParameterDefinition.Flag.FlagOptional
                ):
                    valid = False
                    break

            if valid:
                alg.addParameterSources(param.name(), val)

        outputs = {}
        for output in self._alg.destinationParameterDefinitions():
            if not output.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                wrapper = self.wrappers[output.name()]

                if wrapper.isModelOutput():
                    name = wrapper.modelOutputName()
                    if name:
                        # if there was a previous output definition already for this output, we start with it,
                        # otherwise we'll lose any existing output comments, coloring, position, etc
                        model_output = self.previous_output_definitions.get(
                            output.name(), QgsProcessingModelOutput(name, name)
                        )
                        model_output.setDescription(name)
                        model_output.setChildId(alg.childId())
                        model_output.setChildOutputName(output.name())
                        outputs[name] = model_output
                else:
                    val = wrapper.value()

                    if isinstance(val, QgsProcessingModelChildParameterSource):
                        val = [val]

                    alg.addParameterSources(output.name(), val)

            if output.flags() & QgsProcessingParameterDefinition.Flag.FlagIsModelOutput:
                if output.name() not in outputs:
                    model_output = QgsProcessingModelOutput(
                        output.name(), output.name()
                    )
                    model_output.setChildId(alg.childId())
                    model_output.setChildOutputName(output.name())
                    outputs[output.name()] = model_output

        alg.setModelOutputs(outputs)
        alg.setDependencies(self.dependencies_panel.value())

        return alg


class ModelerParametersWidget(QgsProcessingModelConfigWidget):
    def __init__(
        self, alg, model, algName=None, configuration=None, dialog=None, context=None
    ):
        super().__init__()
        self._alg = alg  # The algorithm to define in this dialog. It is an instance of QgsProcessingAlgorithm
        self.model = model  # The model this algorithm is going to be added to. It is an instance of QgsProcessingModelAlgorithm
        self.childId = algName  # The name of the algorithm in the model, in case we are editing it and not defining it for the first time
        self.configuration = configuration
        self.context = context
        self.dialog = dialog

        self.widget = ModelerParametersPanelWidget(
            alg, model, algName, configuration, dialog, context
        )
        self.widget.widgetChanged.connect(self.widgetChanged)

        class ContextGenerator(QgsProcessingContextGenerator):
            def __init__(self, context):
                super().__init__()
                self.processing_context = context

            def processingContext(self):
                return self.processing_context

        self.context_generator = ContextGenerator(self.context)

        self.setupUi()
        self.params = None

    def algorithm(self):
        return self._alg

    def switchToCommentTab(self):
        self.tab.setCurrentIndex(1)
        self.commentEdit.setFocus()
        self.commentEdit.selectAll()

    def setupUi(self):
        self.mainLayout = QVBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.tab = QTabWidget()
        self.mainLayout.addWidget(self.tab)

        self.param_widget = QgsPanelWidgetStack()
        self.widget.setDockMode(True)
        self.param_widget.setMainPanel(self.widget)

        self.tab.addTab(self.param_widget, self.tr("Properties"))

        self.commentLayout = QVBoxLayout()
        self.commentEdit = QTextEdit()
        self.commentEdit.setAcceptRichText(False)
        self.commentLayout.addWidget(self.commentEdit, 1)

        hl = QHBoxLayout()
        hl.setContentsMargins(0, 0, 0, 0)
        hl.addWidget(QLabel(self.tr("Color")))
        self.comment_color_button = QgsColorButton()
        self.comment_color_button.setAllowOpacity(True)
        self.comment_color_button.setWindowTitle(self.tr("Comment Color"))
        self.comment_color_button.setShowNull(True, self.tr("Default"))
        hl.addWidget(self.comment_color_button)
        self.commentLayout.addLayout(hl)

        w2 = QWidget()
        w2.setLayout(self.commentLayout)
        self.tab.addTab(w2, self.tr("Comments"))

        self.setLayout(self.mainLayout)

    def setComments(self, text):
        self.commentEdit.setPlainText(text)

    def comments(self):
        return self.commentEdit.toPlainText()

    def setCommentColor(self, color):
        if color.isValid():
            self.comment_color_button.setColor(color)
        else:
            self.comment_color_button.setToNull()

    def commentColor(self):
        return (
            self.comment_color_button.color()
            if not self.comment_color_button.isNull()
            else QColor()
        )

    def setPreviousValues(self):
        self.widget.setPreviousValues()

    def createAlgorithm(self):
        alg = self.widget.createAlgorithm()
        if alg:
            alg.comment().setDescription(self.comments())
            alg.comment().setColor(self.commentColor())
        return alg



"""
***************************************************************************
    BatchAlgorithmDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import time
from typing import Optional

from qgis.core import (
    QgsProcessingOutputBoolean,
    QgsProcessingOutputHtml,
    QgsProcessingOutputNumber,
    QgsProcessingOutputString,
    QgsProject,
)
from qgis.gui import QgsProcessingBatchAlgorithmDialogBase
from qgis.PyQt.QtWidgets import QMainWindow
from qgis.utils import iface

from processing.core.ProcessingResults import resultsList
from processing.gui.BatchPanel import BatchPanel
from processing.gui.Postprocessing import handleAlgorithmResults
from processing.tools import dataobjects
from processing.tools.system import getTempFilename


class BatchAlgorithmDialog(QgsProcessingBatchAlgorithmDialogBase):
    def __init__(
        self,
        alg,
        parent: Optional[QMainWindow] = None,
    ):
        super().__init__(parent or (iface and iface.mainWindow()))

        self.setAlgorithm(alg)

        self.setWindowTitle(
            self.tr("Batch Processing - {0}").format(self.algorithm().displayName())
        )
        self.setMainWidget(BatchPanel(self, self.algorithm()))

        self.context = None
        self.hideShortHelp()

    def runAsSingle(self):
        self.close()

        alg_instance = self.algorithm().create()
        widget = alg_instance.createCustomParametersWidget(parent=iface.mainWindow())
        if not widget:
            from processing.gui.algorithm_widget import AlgorithmWidget

            widget = AlgorithmWidget(
                alg_instance, parent=(iface and iface.mainWindow())
            )

        widget.exec()

    def processingContext(self):
        if self.context is None:
            self.feedback = self.createFeedback()
            self.context = dataobjects.createContext(self.feedback)
            self.context.setLogLevel(self.logLevel())
        return self.context

    def createContext(self, feedback):
        return dataobjects.createContext(feedback)

    def runAlgorithm(self):
        alg_parameters = []

        load_layers = self.mainWidget().checkLoadLayersOnCompletion.isChecked()
        project = QgsProject.instance() if load_layers else None

        for row in range(self.mainWidget().batchRowCount()):
            parameters, ok = self.mainWidget().parametersForRow(
                row=row,
                context=self.processingContext(),
                destinationProject=project,
                warnOnInvalid=True,
            )
            if ok:
                alg_parameters.append(parameters)
        if not alg_parameters:
            return

        self.execute(alg_parameters)

    def handleAlgorithmResults(self, algorithm, context, feedback, parameters):
        handleAlgorithmResults(algorithm, context, feedback, parameters)

    def loadHtmlResults(self, results, num):
        for out in self.algorithm().outputDefinitions():
            if (
                isinstance(out, QgsProcessingOutputHtml)
                and out.name() in results
                and results[out.name()]
            ):
                resultsList.addResult(
                    icon=self.algorithm().icon(),
                    name=f"{out.description()} [{num}]",
                    result=results[out.name()],
                )

    def createSummaryTable(self, algorithm_results, errors):
        createTable = False

        for out in self.algorithm().outputDefinitions():
            if isinstance(
                out,
                (
                    QgsProcessingOutputNumber,
                    QgsProcessingOutputString,
                    QgsProcessingOutputBoolean,
                ),
            ):
                createTable = True
                break

        if not createTable and not errors:
            return

        outputFile = getTempFilename("html")
        with open(outputFile, "w", encoding="utf-8") as f:
            if createTable:
                for i, res in enumerate(algorithm_results):
                    results = res["results"]
                    params = res["parameters"]
                    if i > 0:
                        f.write("<hr>\n")
                    f.write(self.tr("<h3>Parameters</h3>\n"))
                    f.write("<table>\n")
                    for param in self.algorithm().parameterDefinitions():
                        if not param.isDestination():
                            if param.name() in params:
                                f.write(
                                    f"<tr><th>{param.description()}</th><td>{params[param.name()]}</td></tr>\n"
                                )
                    f.write("</table>\n")
                    f.write(self.tr("<h3>Results</h3>\n"))
                    f.write("<table>\n")
                    for out in self.algorithm().outputDefinitions():
                        if out.name() in results:
                            f.write(
                                f"<tr><th>{out.description()}</th><td>{results[out.name()]}</td></tr>\n"
                            )
                    f.write("</table>\n")
            if errors:
                f.write('<h2 style="color: red">{}</h2>\n'.format(self.tr("Errors")))
            for i, res in enumerate(errors):
                errors = res["errors"]
                params = res["parameters"]
                if i > 0:
                    f.write("<hr>\n")
                f.write(self.tr("<h3>Parameters</h3>\n"))
                f.write("<table>\n")
                for param in self.algorithm().parameterDefinitions():
                    if not param.isDestination():
                        if param.name() in params:
                            f.write(
                                f"<tr><th>{param.description()}</th><td>{params[param.name()]}</td></tr>\n"
                            )
                f.write("</table>\n")
                f.write("<h3>{}</h3>\n".format(self.tr("Error")))
                f.write('<p style="color: red">{}</p>\n'.format("<br>".join(errors)))

        resultsList.addResult(
            icon=self.algorithm().icon(),
            name=f"{self.algorithm().name()} [summary]",
            timestamp=time.localtime(),
            result=outputFile,
        )



"""
***************************************************************************
    common.py
    ---------------------
    Date                 : January 2024
    Copyright            : (C) 2024 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

import functools
from enum import Enum

def _handle_extended_enum_values(v):
    """
    If v is an Enum, returns an integer representation of its value wrapped
    in the Enum class.
    Otherwise returns v unchanged.
    """
    if isinstance(v, Enum):
       # v may be an Enum value which itself is an Enum value!
       # E.g. QNetworkRequest.Attribute(QgsNetworkRequestParameters.RequestAttributes.AttributeInitiatorClass)
       # in this case v is an Enum of the class QNetworkRequest.Attribute, while v.value is an Enum of the
       # class QgsNetworkRequestParameters.RequestAttributes. For this to work, we need to extract the integer
       # value from QgsNetworkRequestParameters.RequestAttributes.AttributeInitiatorClass and then convert it
       # to a QNetworkRequest.Attribute enum value!
       if isinstance(v.value, Enum):
          return v.__class__(int(v.value.value))
       return v.__class__(int(v.value))
    return v

def _extended_enum_wrapper(method):
    """
    Calls a function, first converting all passed Enum arguments to integer
    values so that extended enum values work correctly.
    """
    @functools.wraps(method)
    def wrapped(*args, **kwargs):
        parsed_args = [_handle_extended_enum_values(arg) for arg in args]
        parsed_kwargs = {k: _handle_extended_enum_values(v) for k, v in kwargs.items()}
        return method(*parsed_args, **parsed_kwargs)

    return wrapped


def install_extended_enum_wrapper(f):
    """
    Installs a wrapper on a function so that all passed extended Enum arguments are handled correctly.
    """
    return _extended_enum_wrapper(f)




"""
***************************************************************************
    ModelerUtils.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from processing.core.ProcessingConfig import ProcessingConfig
from processing.tools.system import mkdir, userFolder


class ModelerUtils:
    MODELS_FOLDER = "MODELS_FOLDER"

    @staticmethod
    def defaultModelsFolder():
        folder = str(os.path.join(userFolder(), "models"))
        mkdir(folder)
        return os.path.abspath(folder)

    @staticmethod
    def modelsFolders():
        folder = ProcessingConfig.getSetting(ModelerUtils.MODELS_FOLDER)
        if folder is not None:
            return folder.split(";")
        else:
            return [ModelerUtils.defaultModelsFolder()]



"""
***************************************************************************
    grass_algorithm.py
    ---------------------
    Date                 : February 2015
    Copyright            : (C) 2014-2015 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "February 2015"
__copyright__ = "(C) 2012-2015, Victor Olaya"

import importlib
import math
import os
import sys
import uuid
import warnings
from pathlib import Path
from typing import Optional

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsMapLayer,
    QgsMapLayerType,
    QgsMessageLog,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingContext,
    QgsProcessingException,
    QgsProcessingOutputHtml,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterCrs,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterExtent,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterFile,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterNumber,
    QgsProcessingParameterPoint,
    QgsProcessingParameterRange,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
    QgsProcessingParameterVectorLayer,
    QgsProcessingUtils,
    QgsProviderRegistry,
    QgsRasterLayer,
    QgsVectorFileWriter,
    QgsVectorLayer,
)
from qgis.PyQt.QtCore import QCoreApplication, QUrl
from qgis.utils import iface

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from osgeo import ogr

from processing.core.parameters import getParameterFromString
from processing.core.ProcessingConfig import ProcessingConfig
from processing.tools.system import getTempFilename, isWindows

from grassprovider.grass_utils import GrassUtils
from grassprovider.parsed_description import ParsedDescription

pluginPath = os.path.normpath(
    os.path.join(os.path.split(os.path.dirname(__file__))[0], os.pardir)
)


class GrassAlgorithm(QgsProcessingAlgorithm):
    GRASS_OUTPUT_TYPE_PARAMETER = "GRASS_OUTPUT_TYPE_PARAMETER"
    GRASS_MIN_AREA_PARAMETER = "GRASS_MIN_AREA_PARAMETER"
    GRASS_SNAP_TOLERANCE_PARAMETER = "GRASS_SNAP_TOLERANCE_PARAMETER"
    GRASS_REGION_EXTENT_PARAMETER = "GRASS_REGION_PARAMETER"
    GRASS_REGION_CELLSIZE_PARAMETER = "GRASS_REGION_CELLSIZE_PARAMETER"
    GRASS_REGION_ALIGN_TO_RESOLUTION = "GRASS_REGION_ALIGN_TO_RESOLUTION"
    GRASS_RASTER_FORMAT_OPT = "GRASS_RASTER_FORMAT_OPT"
    GRASS_RASTER_FORMAT_META = "GRASS_RASTER_FORMAT_META"
    GRASS_VECTOR_DSCO = "GRASS_VECTOR_DSCO"
    GRASS_VECTOR_LCO = "GRASS_VECTOR_LCO"
    GRASS_VECTOR_EXPORT_NOCAT = "GRASS_VECTOR_EXPORT_NOCAT"

    OUTPUT_TYPES = ["auto", "point", "line", "area"]
    QGIS_OUTPUT_TYPES = {
        QgsProcessing.SourceType.TypeVectorAnyGeometry: "auto",
        QgsProcessing.SourceType.TypeVectorPoint: "point",
        QgsProcessing.SourceType.TypeVectorLine: "line",
        QgsProcessing.SourceType.TypeVectorPolygon: "area",
    }

    def __init__(
        self,
        description_file: Optional[Path] = None,
        json_definition: Optional[dict] = None,
        description_folder: Optional[Path] = None,
    ):
        super().__init__()
        self._name = ""
        self._display_name = ""
        self._short_description = ""
        self._group = ""
        self._groupId = ""
        self.grass_name = ""
        self.params = []
        self.hardcodedStrings = []
        self.inputLayers = []
        self.commands = []
        self.outputCommands = []
        self.exportedLayers = {}
        self.fileOutputs = {}
        self._description_file: Optional[Path] = description_file
        self._json_definition: Optional[dict] = json_definition
        self._description_folder: Optional[Path] = description_folder

        # Default GRASS parameters
        self.region = None
        self.cellSize = None
        self.snapTolerance = None
        self.outputType = None
        self.minArea = None
        self.alignToResolution = None

        # destination Crs for combineLayerExtents, will be set from layer or mapSettings
        self.destination_crs = QgsCoordinateReferenceSystem()

        # Load parameters from a description file
        if self._description_file is not None:
            self._define_characteristics_from_file()
        else:
            self._define_characteristics_from_json()

        self.numExportedLayers = 0
        # Do we need this anymore?
        self.uniqueSuffix = str(uuid.uuid4()).replace("-", "")

        # Use the ext mechanism
        self.module = None
        try:
            extpath = None
            ext_name = None
            if self._description_file:
                ext_name = self.name().replace(".", "_")
                extpath = self._description_file.parents[1].joinpath(
                    "ext", ext_name + ".py"
                )
            elif self._json_definition.get("ext_path"):
                ext_name = self._json_definition["ext_path"]
                extpath = self._description_folder.parents[0].joinpath(
                    "ext", ext_name + ".py"
                )

            # this check makes it a bit faster
            if extpath and extpath.exists():
                spec = importlib.util.spec_from_file_location(
                    "grassprovider.ext." + ext_name, extpath
                )
                self.module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(self.module)

        except Exception as e:
            QgsMessageLog.logMessage(
                self.tr("Failed to load: {0}\n{1}").format(extpath, e),
                "Processing",
                Qgis.MessageLevel.Critical,
            )
            pass

    def createInstance(self):
        return self.__class__(
            description_file=self._description_file,
            json_definition=self._json_definition,
            description_folder=self._description_folder,
        )

    def name(self):
        return self._name

    def displayName(self):
        return self._display_name

    def shortDescription(self):
        return self._short_description

    def group(self):
        return self._group

    def groupId(self):
        return self._groupId

    def icon(self):
        return QgsApplication.getThemeIcon("/providerGrass.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("providerGrass.svg")

    def flags(self):
        # TODO - maybe it's safe to background thread this?
        return (
            super().flags()
            | QgsProcessingAlgorithm.Flag.FlagNoThreading
            | QgsProcessingAlgorithm.Flag.FlagDisplayNameIsLiteral
        )

    def tr(self, string: str):
        return QCoreApplication.translate(self.__class__.__name__, string)

    def helpUrl(self):
        helpPath = GrassUtils.grassHelpPath()
        if helpPath == "":
            return None

        if os.path.exists(helpPath):
            return QUrl.fromLocalFile(
                os.path.join(helpPath, f"{self.grass_name}.html")
            ).toString()
        else:
            return helpPath + f"{self.grass_name}.html"

    def initAlgorithm(self, config=None):
        """
        Algorithm initialization
        """
        for p in self.params:
            # We use createOutput argument for automatic output creation
            self.addParameter(p, True)

    def _define_characteristics_from_file(self):
        """
        Create algorithm parameters and outputs from a text file.
        """
        results = ParsedDescription.parse_description_file(self._description_file)
        self._define_characteristics_from_parsed_description(results)

    def _define_characteristics_from_json(self):
        """
        Create algorithm parameters and outputs from JSON definition.
        """
        results = ParsedDescription.from_dict(self._json_definition)
        self._define_characteristics_from_parsed_description(results)

    def _define_characteristics_from_parsed_description(
        self, description: ParsedDescription
    ):
        """
        Create algorithm parameters and outputs from parsed description
        """
        self.grass_name = description.grass_command
        self._name = description.name
        self._short_description = description.short_description
        self._display_name = description.display_name
        self._group = description.group
        self._groupId = description.group_id
        self.hardcodedStrings = description.hardcoded_strings[:]

        self.params = []

        has_raster_input: bool = False
        has_vector_input: bool = False

        has_raster_output: bool = False
        has_vector_outputs: bool = False

        for param_string in description.param_strings:
            try:
                parameter = getParameterFromString(param_string, "GrassAlgorithm")
            except Exception as e:
                QgsMessageLog.logMessage(
                    self.tr("Could not open GRASS algorithm: {0}").format(self._name),
                    self.tr("Processing"),
                    Qgis.MessageLevel.Critical,
                )
                raise e

            if parameter is None:
                continue

            self.params.append(parameter)
            if isinstance(
                parameter,
                (
                    QgsProcessingParameterVectorLayer,
                    QgsProcessingParameterFeatureSource,
                ),
            ):
                has_vector_input = True
            elif isinstance(parameter, QgsProcessingParameterRasterLayer):
                has_raster_input = True
            elif isinstance(parameter, QgsProcessingParameterMultipleLayers):
                if parameter.layerType() < 3 or parameter.layerType() == 5:
                    has_vector_input = True
                elif parameter.layerType() == 3:
                    has_raster_input = True
            elif isinstance(parameter, QgsProcessingParameterVectorDestination):
                has_vector_outputs = True
            elif isinstance(parameter, QgsProcessingParameterRasterDestination):
                has_raster_output = True

        param = QgsProcessingParameterExtent(
            self.GRASS_REGION_EXTENT_PARAMETER,
            self.tr("GRASS region extent"),
            optional=True,
        )
        param.setFlags(
            param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.params.append(param)

        if has_raster_output or has_raster_input:
            # Add a cellsize parameter
            param = QgsProcessingParameterNumber(
                self.GRASS_REGION_CELLSIZE_PARAMETER,
                self.tr("GRASS region cellsize (leave 0 for default)"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                maxValue=sys.float_info.max + 1,
                defaultValue=0.0,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

        if has_raster_output:
            # Add a createopt parameter for format export
            param = QgsProcessingParameterString(
                self.GRASS_RASTER_FORMAT_OPT,
                self.tr("Output Rasters format options (createopt)"),
                multiLine=True,
                optional=True,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            param.setHelp(self.tr("Creation options should be comma separated"))
            self.params.append(param)

            # Add a metadata parameter for format export
            param = QgsProcessingParameterString(
                self.GRASS_RASTER_FORMAT_META,
                self.tr("Output Rasters format metadata options (metaopt)"),
                multiLine=True,
                optional=True,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            param.setHelp(self.tr("Metadata options should be comma separated"))
            self.params.append(param)

        if has_vector_input:
            param = QgsProcessingParameterNumber(
                self.GRASS_SNAP_TOLERANCE_PARAMETER,
                self.tr("v.in.ogr snap tolerance (-1 = no snap)"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=-1.0,
                maxValue=sys.float_info.max + 1,
                defaultValue=-1.0,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)
            param = QgsProcessingParameterNumber(
                self.GRASS_MIN_AREA_PARAMETER,
                self.tr("v.in.ogr min area"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                maxValue=sys.float_info.max + 1,
                defaultValue=0.0001,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

        if has_vector_outputs:
            # Add an optional output type
            param = QgsProcessingParameterEnum(
                self.GRASS_OUTPUT_TYPE_PARAMETER,
                self.tr("v.out.ogr output type"),
                self.OUTPUT_TYPES,
                defaultValue=0,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

            # Add a DSCO parameter for format export
            param = QgsProcessingParameterString(
                self.GRASS_VECTOR_DSCO,
                self.tr("v.out.ogr output data source options (dsco)"),
                multiLine=True,
                optional=True,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

            # Add a LCO parameter for format export
            param = QgsProcessingParameterString(
                self.GRASS_VECTOR_LCO,
                self.tr("v.out.ogr output layer options (lco)"),
                multiLine=True,
                optional=True,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

            # Add a -c flag for export
            param = QgsProcessingParameterBoolean(
                self.GRASS_VECTOR_EXPORT_NOCAT,
                self.tr(
                    "Also export features without category (not labeled). Otherwise only features with category are exported"
                ),
                False,
            )
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.params.append(param)

    def getDefaultCellSize(self):
        """
        Determine a default cell size from all the raster layers.
        """
        cellsize = 0.0
        layers = [l for l in self.inputLayers if isinstance(l, QgsRasterLayer)]

        for layer in layers:
            cellsize = max(layer.rasterUnitsPerPixelX(), cellsize)

        if cellsize == 0.0:
            cellsize = 100.0

        return cellsize

    def grabDefaultGrassParameters(self, parameters, context):
        """
        Imports default GRASS parameters (EXTENT, etc) into
        object attributes for faster retrieving.
        """
        # GRASS region extent
        self.region = self.parameterAsExtent(
            parameters, self.GRASS_REGION_EXTENT_PARAMETER, context
        )
        # GRASS cell size
        if self.parameterDefinition(self.GRASS_REGION_CELLSIZE_PARAMETER):
            self.cellSize = self.parameterAsDouble(
                parameters, self.GRASS_REGION_CELLSIZE_PARAMETER, context
            )
        # GRASS snap tolerance
        self.snapTolerance = self.parameterAsDouble(
            parameters, self.GRASS_SNAP_TOLERANCE_PARAMETER, context
        )
        # GRASS min area
        self.minArea = self.parameterAsDouble(
            parameters, self.GRASS_MIN_AREA_PARAMETER, context
        )
        # GRASS output type
        self.outputType = self.parameterAsString(
            parameters, self.GRASS_OUTPUT_TYPE_PARAMETER, context
        )
        # GRASS align to resolution
        self.alignToResolution = self.parameterAsBoolean(
            parameters, self.GRASS_REGION_ALIGN_TO_RESOLUTION, context
        )

    def processAlgorithm(self, original_parameters, context, feedback):
        if isWindows():
            path = GrassUtils.grassPath()
            if path == "":
                raise QgsProcessingException(
                    self.tr(
                        "GRASS folder is not configured. Please "
                        "configure it before running GRASS algorithms."
                    )
                )

        version = GrassUtils.installedVersion(True)
        if version is None:
            raise QgsProcessingException(
                self.tr(
                    "Problem with GRASS installation: GRASS was not found or is not correctly installed"
                )
            )

        # make a copy of the original parameters dictionary - it gets modified by grass algorithms
        parameters = {k: v for k, v in original_parameters.items()}

        # Create brand new commands lists
        self.commands = []
        self.outputCommands = []
        self.exportedLayers = {}
        self.fileOutputs = {}

        # If GRASS session has been created outside of this algorithm then
        # get the list of layers loaded in GRASS otherwise start a new
        # session
        existingSession = GrassUtils.sessionRunning
        if existingSession:
            self.exportedLayers = GrassUtils.getSessionLayers()
        else:
            GrassUtils.startGrassSession()

        # Handle default GRASS parameters
        self.grabDefaultGrassParameters(parameters, context)

        # Handle ext functions for inputs/command/outputs
        for fName in ["Inputs", "Command", "Outputs"]:
            fullName = f"process{fName}"
            if self.module and hasattr(self.module, fullName):
                getattr(self.module, fullName)(self, parameters, context, feedback)
            else:
                getattr(self, fullName)(parameters, context, feedback)

        # Run GRASS
        loglines = [self.tr("GRASS execution commands")]
        for line in self.commands:
            feedback.pushCommandInfo(line)
            loglines.append(line)
        if ProcessingConfig.getSetting(GrassUtils.GRASS_LOG_COMMANDS):
            QgsMessageLog.logMessage(
                "\n".join(loglines), self.tr("Processing"), Qgis.MessageLevel.Info
            )

        GrassUtils.executeGrass(self.commands, feedback, self.outputCommands)

        # If the session has been created outside of this algorithm, add
        # the new GRASS layers to it otherwise finish the session
        if existingSession:
            GrassUtils.addSessionLayers(self.exportedLayers)
        else:
            GrassUtils.endGrassSession()

        # Return outputs map
        outputs = {}
        for out in self.outputDefinitions():
            outName = out.name()
            if outName in parameters:
                if outName in self.fileOutputs:
                    outputs[outName] = self.fileOutputs[outName]
                else:
                    outputs[outName] = parameters[outName]
                if isinstance(out, QgsProcessingOutputHtml):
                    if self.module and hasattr(self.module, "convertToHtml"):
                        func = getattr(self.module, "convertToHtml")
                        func(self, self.fileOutputs[outName], outputs)
                    else:
                        self.convertToHtml(self.fileOutputs[outName])
        return outputs

    def processInputs(self, parameters, context, feedback):
        """Prepare the GRASS import commands"""
        inputs = [
            p
            for p in self.parameterDefinitions()
            if isinstance(
                p,
                (
                    QgsProcessingParameterVectorLayer,
                    QgsProcessingParameterFeatureSource,
                    QgsProcessingParameterRasterLayer,
                    QgsProcessingParameterMultipleLayers,
                ),
            )
        ]
        for param in inputs:
            paramName = param.name()
            if paramName not in parameters:
                continue
            # Handle Null parameter
            if parameters[paramName] is None:
                continue
            elif (
                isinstance(parameters[paramName], str)
                and len(parameters[paramName]) == 0
            ):
                continue

            # Raster inputs needs to be imported into temp GRASS DB
            if isinstance(param, QgsProcessingParameterRasterLayer):
                if paramName not in self.exportedLayers:
                    self.loadRasterLayerFromParameter(paramName, parameters, context)
            # Vector inputs needs to be imported into temp GRASS DB
            elif isinstance(
                param,
                (
                    QgsProcessingParameterFeatureSource,
                    QgsProcessingParameterVectorLayer,
                ),
            ):
                if paramName not in self.exportedLayers:
                    # Attribute tables are also vector inputs
                    if QgsProcessing.SourceType.TypeFile in param.dataTypes():
                        self.loadAttributeTableFromParameter(
                            paramName, parameters, context
                        )
                    else:
                        self.loadVectorLayerFromParameter(
                            paramName,
                            parameters,
                            context,
                            external=None,
                            feedback=feedback,
                        )
            # For multiple inputs, process each layer
            elif isinstance(param, QgsProcessingParameterMultipleLayers):
                layers = self.parameterAsLayerList(parameters, paramName, context)
                for idx, layer in enumerate(layers):
                    layerName = f"{paramName}_{idx}"
                    # Add a raster layer
                    if layer.type() == QgsMapLayerType.RasterLayer:
                        self.loadRasterLayer(layerName, layer, context)
                    # Add a vector layer
                    elif layer.type() == QgsMapLayerType.VectorLayer:
                        self.loadVectorLayer(
                            layerName, layer, context, external=None, feedback=feedback
                        )

        self.postInputs(context)

    def postInputs(self, context: QgsProcessingContext):
        """
        After layer imports, we need to update some internal parameters
        """
        # If projection has not already be set, use the project
        self.setSessionProjectionFromProject(context)

        # Build GRASS region
        if self.region.isEmpty():
            self.region = QgsProcessingUtils.combineLayerExtents(
                self.inputLayers, self.destination_crs, context
            )
        command = f"g.region n={self.region.yMaximum()} s={self.region.yMinimum()} e={self.region.xMaximum()} w={self.region.xMinimum()}"
        # Handle cell size
        if self.parameterDefinition(self.GRASS_REGION_CELLSIZE_PARAMETER):
            if self.cellSize:
                cellSize = self.cellSize
            else:
                cellSize = self.getDefaultCellSize()
            command += f" res={cellSize}"

        # Handle align to resolution
        if self.alignToResolution:
            command += " -a"

        # Add the default parameters commands
        self.commands.append(command)

        QgsMessageLog.logMessage(
            self.tr("processInputs end. Commands: {}").format(self.commands),
            "Grass",
            Qgis.MessageLevel.Info,
        )

    def processCommand(self, parameters, context, feedback, delOutputs=False):
        """
        Prepare the GRASS algorithm command
        :param parameters:
        :param context:
        :param delOutputs: do not add outputs to commands.
        """
        noOutputs = [
            o
            for o in self.parameterDefinitions()
            if o not in self.destinationParameterDefinitions()
        ]
        command = f"{self.grass_name} "
        command += "{}".join(self.hardcodedStrings)

        # Add algorithm command
        for param in noOutputs:
            paramName = param.name()
            value = None

            # Exclude default GRASS parameters
            if paramName in [
                self.GRASS_REGION_CELLSIZE_PARAMETER,
                self.GRASS_REGION_EXTENT_PARAMETER,
                self.GRASS_MIN_AREA_PARAMETER,
                self.GRASS_SNAP_TOLERANCE_PARAMETER,
                self.GRASS_OUTPUT_TYPE_PARAMETER,
                self.GRASS_REGION_ALIGN_TO_RESOLUTION,
                self.GRASS_RASTER_FORMAT_OPT,
                self.GRASS_RASTER_FORMAT_META,
                self.GRASS_VECTOR_DSCO,
                self.GRASS_VECTOR_LCO,
                self.GRASS_VECTOR_EXPORT_NOCAT,
            ]:
                continue

            # Raster and vector layers
            if isinstance(
                param,
                (
                    QgsProcessingParameterRasterLayer,
                    QgsProcessingParameterVectorLayer,
                    QgsProcessingParameterFeatureSource,
                ),
            ):
                if paramName in self.exportedLayers:
                    value = self.exportedLayers[paramName]
                else:
                    value = self.parameterAsCompatibleSourceLayerPath(
                        parameters,
                        paramName,
                        context,
                        QgsVectorFileWriter.supportedFormatExtensions(),
                    )
            # MultipleLayers
            elif isinstance(param, QgsProcessingParameterMultipleLayers):
                layers = self.parameterAsLayerList(parameters, paramName, context)
                values = []
                for idx in range(len(layers)):
                    layerName = f"{paramName}_{idx}"
                    values.append(self.exportedLayers[layerName])
                value = ",".join(values)
            # For booleans, we just add the parameter name
            elif isinstance(param, QgsProcessingParameterBoolean):
                if self.parameterAsBoolean(parameters, paramName, context):
                    command += f" {paramName}"
            # For Extents, remove if the value is null
            elif isinstance(param, QgsProcessingParameterExtent):
                if self.parameterAsExtent(parameters, paramName, context):
                    value = self.parameterAsString(parameters, paramName, context)
            # For enumeration, we need to grab the string value
            elif isinstance(param, QgsProcessingParameterEnum):
                # Handle multiple values
                if param.allowMultiple():
                    indexes = self.parameterAsEnums(parameters, paramName, context)
                else:
                    indexes = [self.parameterAsEnum(parameters, paramName, context)]
                if indexes:
                    value = '"{}"'.format(
                        ",".join([param.options()[i] for i in indexes])
                    )
            # For strings, we just translate as string
            elif isinstance(param, QgsProcessingParameterString):
                data = self.parameterAsString(parameters, paramName, context)
                # if string is empty, we don't add it
                if len(data) > 0:
                    value = (
                        f'"{self.parameterAsString(parameters, paramName, context)}"'
                    )
            # For fields, we just translate as string
            elif isinstance(param, QgsProcessingParameterField):
                value = ",".join(self.parameterAsFields(parameters, paramName, context))
            elif isinstance(param, QgsProcessingParameterFile):
                if self.parameterAsString(parameters, paramName, context):
                    value = (
                        f'"{self.parameterAsString(parameters, paramName, context)}"'
                    )
            elif isinstance(param, QgsProcessingParameterPoint):
                if self.parameterAsString(parameters, paramName, context):
                    # parameter specified, evaluate as point
                    # TODO - handle CRS transform
                    point = self.parameterAsPoint(parameters, paramName, context)
                    value = f"{point.x()},{point.y()}"
            # For numbers, we translate as a string
            elif isinstance(
                param, (QgsProcessingParameterNumber, QgsProcessingParameterPoint)
            ):
                value = self.parameterAsString(parameters, paramName, context)
            elif isinstance(param, QgsProcessingParameterRange):
                v = self.parameterAsRange(parameters, paramName, context)
                if (
                    param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
                ) and (math.isnan(v[0]) or math.isnan(v[1])):
                    continue
                else:
                    value = f"{v[0]},{v[1]}"
            elif isinstance(param, QgsProcessingParameterCrs):
                if self.parameterAsCrs(parameters, paramName, context):
                    # TODO: ideally we should be exporting to WKT here, but it seems not all grass algorithms
                    # will accept a wkt string for a crs value (e.g. r.tileset)
                    value = f'"{self.parameterAsCrs(parameters, paramName, context).toProj()}"'
            # For everything else, we assume that it is a string
            else:
                value = f'"{self.parameterAsString(parameters, paramName, context)}"'
            if value:
                command += " {}={}".format(paramName.replace("~", ""), value)

        # Handle outputs
        if not delOutputs:
            for out in self.destinationParameterDefinitions():
                # We exclude hidden parameters
                if out.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                    continue
                outName = out.name()
                # For File destination
                if isinstance(out, QgsProcessingParameterFileDestination):
                    if outName in parameters and parameters[outName] is not None:
                        outPath = self.parameterAsFileOutput(
                            parameters, outName, context
                        )
                        self.fileOutputs[outName] = outPath
                        # for HTML reports, we need to redirect stdout
                        if out.defaultFileExtension().lower() == "html":
                            if outName == "html":
                                # for "fake" outputs redirect command stdout
                                command += f' > "{outPath}"'
                            else:
                                # for real outputs only output itself should be redirected
                                command += f' {outName}=- > "{outPath}"'
                        else:
                            command += f' {outName}="{outPath}"'
                # For folders destination
                elif isinstance(out, QgsProcessingParameterFolderDestination):
                    # We need to add a unique temporary basename
                    uniqueBasename = outName + self.uniqueSuffix
                    command += f" {outName}={uniqueBasename}"
                else:
                    if outName in parameters and parameters[outName] is not None:
                        # We add an output name to make sure it is unique if the session
                        # uses this algorithm several times.
                        uniqueOutputName = outName + self.uniqueSuffix
                        command += f" {outName}={uniqueOutputName}"

                        # Add output file to exported layers, to indicate that
                        # they are present in GRASS
                        self.exportedLayers[outName] = uniqueOutputName

        command += " --overwrite"
        self.commands.append(command)
        QgsMessageLog.logMessage(
            self.tr("processCommands end. Commands: {}").format(self.commands),
            "Grass",
            Qgis.MessageLevel.Info,
        )

    def vectorOutputType(self, parameters, context):
        """Determine vector output types for outputs"""
        self.outType = "auto"
        if self.parameterDefinition(self.GRASS_OUTPUT_TYPE_PARAMETER):
            typeidx = self.parameterAsEnum(
                parameters, self.GRASS_OUTPUT_TYPE_PARAMETER, context
            )
            self.outType = "auto" if typeidx is None else self.OUTPUT_TYPES[typeidx]

    def processOutputs(self, parameters, context, feedback):
        """Prepare the GRASS v.out.ogr commands"""
        # Determine general vector output type
        self.vectorOutputType(parameters, context)

        for out in self.destinationParameterDefinitions():
            outName = out.name()
            if outName not in parameters:
                # skipped output
                continue

            if isinstance(out, QgsProcessingParameterRasterDestination):
                self.exportRasterLayerFromParameter(outName, parameters, context)
            elif isinstance(out, QgsProcessingParameterVectorDestination):
                self.exportVectorLayerFromParameter(outName, parameters, context)
            elif isinstance(out, QgsProcessingParameterFolderDestination):
                self.exportRasterLayersIntoDirectory(outName, parameters, context)

    def loadRasterLayerFromParameter(
        self, name, parameters, context: QgsProcessingContext, external=None, band=1
    ):
        """
        Creates a dedicated command to load a raster into
        the temporary GRASS DB.
        :param name: name of the parameter.
        :param parameters: algorithm parameters dict.
        :param context: algorithm context.
        :param external: use r.external if True, r.in.gdal otherwise.
        :param band: imports only specified band. None for all bands.
        """
        layer = self.parameterAsRasterLayer(parameters, name, context)
        self.loadRasterLayer(name, layer, context, external, band)

    def loadRasterLayer(
        self,
        name,
        layer,
        context: QgsProcessingContext,
        external=None,
        band=1,
        destName=None,
    ):
        """
        Creates a dedicated command to load a raster into
        the temporary GRASS DB.
        :param name: name of the parameter.
        :param layer: QgsMapLayer for the raster layer.
        :param context: Processing context
        :param external: use r.external if True, r.in.gdal if False.
        :param band: imports only specified band. None for all bands.
        :param destName: force the destination name of the raster.
        """
        if external is None:
            external = ProcessingConfig.getSetting(GrassUtils.GRASS_USE_REXTERNAL)
        self.inputLayers.append(layer)
        self.setSessionProjectionFromLayer(layer, context)
        if not destName:
            destName = f"rast_{os.path.basename(getTempFilename(context=context))}"
        self.exportedLayers[name] = destName

        if layer.providerType() == "postgresraster":
            source = ""
            uri = layer.dataProvider().uri()
            source = f"PG: {uri.connectionInfo()}"
            schema = uri.schema()
            if schema:
                source += f" schema='{schema}'"
            table = uri.table()
            source += f" table='{table}'"
            column = uri.param("column") or uri.geometryColumn()
            if column:
                source += f" column='{column}'"
            is_tiled = any(
                [
                    layer.dataProvider().xSize() != layer.dataProvider().xBlockSize(),
                    layer.dataProvider().ySize() != layer.dataProvider().yBlockSize(),
                ]
            )
            source += f" mode={2 if is_tiled else 1}"
            where = layer.dataProvider().subsetString()
            if where:
                source += f" where='{where}'"
        else:
            source = os.path.normpath(layer.source())

        command = '{} input="{}" {}output="{}" --overwrite -o'.format(
            "r.external" if external else "r.in.gdal",
            source,
            f"band={band} " if band else "",
            destName,
        )
        self.commands.append(command)

    def exportRasterLayerFromParameter(
        self, name, parameters, context, colorTable=True
    ):
        """
        Creates a dedicated command to export a raster from
        temporary GRASS DB into a file via gdal.
        :param name: name of the parameter.
        :param parameters: Algorithm parameters dict.
        :param context: Algorithm context.
        :param colorTable: preserve color Table.
        """
        fileName = self.parameterAsOutputLayer(parameters, name, context)
        if not fileName:
            return

        fileName = os.path.normpath(fileName)
        grassName = f"{name}{self.uniqueSuffix}"
        outFormat = GrassUtils.getRasterFormatFromFilename(fileName)
        createOpt = self.parameterAsString(
            parameters, self.GRASS_RASTER_FORMAT_OPT, context
        )
        metaOpt = self.parameterAsString(
            parameters, self.GRASS_RASTER_FORMAT_META, context
        )
        self.exportRasterLayer(
            grassName, fileName, colorTable, outFormat, createOpt, metaOpt
        )

        self.fileOutputs[name] = fileName

    def exportRasterLayer(
        self,
        grassName,
        fileName,
        colorTable=True,
        outFormat="GTiff",
        createOpt=None,
        metaOpt=None,
    ):
        """
        Creates a dedicated command to export a raster from
        temporary GRASS DB into a file via gdal.
        :param grassName: name of the raster to export.
        :param fileName: file path of raster layer.
        :param colorTable: preserve color Table.
        :param outFormat: file format for export.
        :param createOpt: creation options for format.
        :param metaOpt: metadata options for export.
        """
        createOpt = createOpt or GrassUtils.GRASS_RASTER_FORMATS_CREATEOPTS.get(
            outFormat
        )

        for cmd in [self.commands, self.outputCommands]:
            # Adjust region to layer before exporting
            cmd.append(f"g.region raster={grassName}")
            cmd.append(
                'r.out.gdal -t -m{} input="{}" output="{}" format="{}" {}{} --overwrite'.format(
                    "" if colorTable else " -c",
                    grassName,
                    fileName,
                    outFormat,
                    f' createopt="{createOpt}"' if createOpt else "",
                    f' metaopt="{metaOpt}"' if metaOpt else "",
                )
            )

    def exportRasterLayersIntoDirectory(
        self, name, parameters, context, colorTable=True, wholeDB=False
    ):
        """
        Creates a dedicated loop command to export rasters from
        temporary GRASS DB into a directory via gdal.
        :param name: name of the output directory parameter.
        :param parameters: Algorithm parameters dict.
        :param context: Algorithm context.
        :param colorTable: preserve color Table.
        :param wholeDB: export every raster layer from the GRASSDB
        """
        # Grab directory name and temporary basename
        outDir = os.path.normpath(self.parameterAsString(parameters, name, context))
        basename = ""
        if not wholeDB:
            basename = name + self.uniqueSuffix

        # Add a loop export from the basename
        for cmd in [self.commands, self.outputCommands]:
            # TODO Format/options support
            if isWindows():
                cmd.append(f"if not exist {outDir} mkdir {outDir}")
                cmd.append(
                    "for /F %%r IN ('g.list type^=rast pattern^=\"{}*\"') do r.out.gdal -m{} input=%%r output={}/%%r.tif {}".format(
                        basename,
                        " -t" if colorTable else "",
                        outDir,
                        '--overwrite -c createopt="TFW=YES,COMPRESS=LZW"',
                    )
                )
            else:
                cmd.append(f"for r in $(g.list type=rast pattern='{basename}*'); do")
                cmd.append(
                    "  r.out.gdal -m{0} input=${{r}} output={1}/${{r}}.tif {2}".format(
                        " -t" if colorTable else "",
                        outDir,
                        '--overwrite -c createopt="TFW=YES,COMPRESS=LZW"',
                    )
                )
                cmd.append("done")

    def loadVectorLayerFromParameter(
        self, name, parameters, context, feedback, external=False
    ):
        """
        Creates a dedicated command to load a vector into
        the temporary GRASS DB.
        :param name: name of the parameter
        :param parameters: Parameters of the algorithm.
        :param context: Processing context
        :param external: use v.external (v.in.ogr if False).
        """
        layer = self.parameterAsVectorLayer(parameters, name, context)

        is_ogr_disk_based_layer = (
            layer is not None and layer.dataProvider().name() == "ogr"
        )
        if is_ogr_disk_based_layer:
            # we only support direct reading of disk based ogr layers -- not ogr postgres layers, etc
            source_parts = QgsProviderRegistry.instance().decodeUri(
                "ogr", layer.source()
            )
            if not source_parts.get("path"):
                is_ogr_disk_based_layer = False
            elif source_parts.get("layerId"):
                # no support for directly reading layers by id in grass
                is_ogr_disk_based_layer = False

        if not is_ogr_disk_based_layer:
            # parameter is not a vector layer or not an OGR layer - try to convert to a source compatible with
            # grass OGR inputs and extract selection if required
            path = self.parameterAsCompatibleSourceLayerPath(
                parameters,
                name,
                context,
                QgsVectorFileWriter.supportedFormatExtensions(),
                feedback=feedback,
            )
            ogr_layer = QgsVectorLayer(path, "", "ogr")
            self.loadVectorLayer(
                name, ogr_layer, context, external=external, feedback=feedback
            )
        else:
            # already an ogr disk based layer source
            self.loadVectorLayer(
                name, layer, context, external=external, feedback=feedback
            )

    def loadVectorLayer(
        self, name, layer, context: QgsProcessingContext, external=False, feedback=None
    ):
        """
        Creates a dedicated command to load a vector into
        temporary GRASS DB.
        :param name: name of the parameter
        :param layer: QgsMapLayer for the vector layer.
        :param context: Processing context
        :param external: use v.external (v.in.ogr if False).
        :param feedback: feedback object
        """
        # TODO: support multiple input formats
        if external is None:
            external = ProcessingConfig.getSetting(GrassUtils.GRASS_USE_VEXTERNAL)

        source_parts = QgsProviderRegistry.instance().decodeUri("ogr", layer.source())
        file_path = source_parts.get("path")
        layer_name = source_parts.get("layerName")

        # safety check: we can only use external for ogr layers which support random read
        if external:
            if feedback is not None:
                feedback.pushInfo(
                    self.tr("Attempting to use v.external for direct layer read")
                )
            ds = ogr.Open(file_path)
            if ds is not None:
                ogr_layer = ds.GetLayer()
                if ogr_layer is None or not ogr_layer.TestCapability(ogr.OLCRandomRead):
                    if feedback is not None:
                        feedback.reportError(
                            self.tr(
                                "Cannot use v.external: layer does not support random read"
                            )
                        )
                    external = False
            else:
                if feedback is not None:
                    feedback.reportError(
                        self.tr("Cannot use v.external: error reading layer")
                    )
                external = False

        self.inputLayers.append(layer)
        self.setSessionProjectionFromLayer(layer, context)
        destFilename = f"vector_{os.path.basename(getTempFilename(context=context))}"
        self.exportedLayers[name] = destFilename
        command = '{}{}{} input="{}"{} output="{}" --overwrite -o'.format(
            "v.external" if external else "v.in.ogr",
            f" min_area={self.minArea}" if not external else "",
            f" snap={self.snapTolerance}" if not external else "",
            os.path.normpath(file_path),
            f' layer="{layer_name}"' if layer_name else "",
            destFilename,
        )
        if layer.subsetString():
            escaped_subset = layer.subsetString().replace('"', '\\"')
            command += f' where="{escaped_subset}"'

        self.commands.append(command)

    def exportVectorLayerFromParameter(
        self, name, parameters, context, layer=None, nocats=False
    ):
        """
        Creates a dedicated command to export a vector from
        a QgsProcessingParameter.
        :param name: name of the parameter.
        :param context: parameters context.
        :param layer: for vector with multiples layers, exports only one layer.
        :param nocats: do not export GRASS categories.
        """
        fileName = os.path.normpath(
            self.parameterAsOutputLayer(parameters, name, context)
        )
        grassName = f"{name}{self.uniqueSuffix}"

        # Find if there is a dataType
        dataType = self.outType
        if self.outType == "auto":
            parameter = self.parameterDefinition(name)
            if parameter:
                layerType = parameter.dataType()
                if layerType in self.QGIS_OUTPUT_TYPES:
                    dataType = self.QGIS_OUTPUT_TYPES[layerType]

        outFormat = QgsVectorFileWriter.driverForExtension(
            os.path.splitext(fileName)[1]
        ).replace(" ", "_")
        dsco = self.parameterAsString(parameters, self.GRASS_VECTOR_DSCO, context)
        lco = self.parameterAsString(parameters, self.GRASS_VECTOR_LCO, context)
        exportnocat = self.parameterAsBoolean(
            parameters, self.GRASS_VECTOR_EXPORT_NOCAT, context
        )
        self.exportVectorLayer(
            grassName,
            fileName,
            layer,
            nocats,
            dataType,
            outFormat,
            dsco,
            lco,
            exportnocat,
        )

        self.fileOutputs[name] = fileName

    def exportVectorLayer(
        self,
        grassName,
        fileName,
        layer=None,
        nocats=False,
        dataType="auto",
        outFormat=None,
        dsco=None,
        lco=None,
        exportnocat=False,
    ):
        """
        Creates a dedicated command to export a vector from
        temporary GRASS DB into a file via OGR.
        :param grassName: name of the vector to export.
        :param fileName: file path of vector layer.
        :param dataType: export only this type of data.
        :param layer: for vector with multiples layers, exports only one layer.
        :param nocats: do not export GRASS categories.
        :param outFormat: file format for export.
        :param dsco: datasource creation options for format.
        :param lco: layer creation options for format.
        :param exportnocat: do not export features without categories.
        """
        if outFormat is None:
            outFormat = QgsVectorFileWriter.driverForExtension(
                os.path.splitext(fileName)[1]
            ).replace(" ", "_")

        for cmd in [self.commands, self.outputCommands]:
            cmd.append(
                'v.out.ogr{} type="{}" input="{}" output="{}" format="{}" {}{}{}{} --overwrite'.format(
                    "" if nocats else "",
                    dataType,
                    grassName,
                    fileName,
                    outFormat,
                    f"layer={layer}" if layer else "",
                    f' dsco="{dsco}"' if dsco else "",
                    f' lco="{lco}"' if lco else "",
                    " -c" if exportnocat else "",
                )
            )

    def loadAttributeTableFromParameter(self, name, parameters, context):
        """
        Creates a dedicated command to load an attribute table
        into the temporary GRASS DB.
        :param name: name of the parameter
        :param parameters: Parameters of the algorithm.
        :param context: Processing context
        """
        table = self.parameterAsVectorLayer(parameters, name, context)
        self.loadAttributeTable(name, table, context)

    def loadAttributeTable(
        self, name, layer, context: QgsProcessingContext, destName=None
    ):
        """
        Creates a dedicated command to load an attribute table
        into the temporary GRASS DB.
        :param name: name of the input parameter.
        :param layer: a layer object to import from.
        :param context: processing context
        :param destName: force the name for the table into GRASS DB.
        """
        self.inputLayers.append(layer)
        if not destName:
            destName = f"table_{os.path.basename(getTempFilename(context=context))}"
        self.exportedLayers[name] = destName
        command = f'db.in.ogr --overwrite input="{os.path.normpath(layer.source())}" output="{destName}"'
        self.commands.append(command)

    def exportAttributeTable(self, grassName, fileName, outFormat="CSV", layer=1):
        """
        Creates a dedicated command to export an attribute
        table from the temporary GRASS DB into a file via ogr.
        :param grassName: name of the parameter.
        :param fileName: file path of raster layer.
        :param outFormat: file format for export.
        :param layer: In GRASS a vector can have multiple layers.
        """
        for cmd in [self.commands, self.outputCommands]:
            cmd.append(
                f'db.out.ogr input="{grassName}" output="{fileName}" layer={layer} format={outFormat} --overwrite'
            )

    def setSessionProjectionFromProject(self, context: QgsProcessingContext):
        """
        Set the projection from the project.
        We create a WKT definition which is transmitted to Grass
        """
        if not GrassUtils.projectionSet and iface:
            self.setSessionProjection(
                iface.mapCanvas().mapSettings().destinationCrs(), context
            )

    def setSessionProjectionFromLayer(
        self, layer: QgsMapLayer, context: QgsProcessingContext
    ):
        """
        Set the projection from a QgsVectorLayer.
        We create a WKT definition which is transmitted to Grass
        """
        if not GrassUtils.projectionSet:
            self.setSessionProjection(layer.crs(), context)

    def setSessionProjection(self, crs, context: QgsProcessingContext):
        """
        Set the session projection to the specified CRS
        """
        self.destination_crs = crs
        file_name = GrassUtils.exportCrsWktToFile(crs, context)
        command = f'g.proj -c wkt="{file_name}"'
        self.commands.append(command)
        GrassUtils.projectionSet = True

    def convertToHtml(self, fileName):
        # Read HTML contents
        lines = []
        with open(fileName, encoding="utf-8") as f:
            lines = f.readlines()

        if len(lines) > 1 and "<html>" not in lines[0]:
            # Then write into the HTML file
            with open(fileName, "w", encoding="utf-8") as f:
                f.write("<html><head>")
                f.write(
                    '<meta http-equiv="Content-Type" content="text/html; charset=utf-8" /></head>'
                )
                f.write("<body><p>")
                for line in lines:
                    f.write(f"{line}</br>")
                f.write("</p></body></html>")

    def canExecute(self):
        message = GrassUtils.checkGrassIsInstalled()
        return not message, message

    def checkParameterValues(self, parameters, context):
        grass_parameters = {k: v for k, v in parameters.items()}
        if self.module:
            if hasattr(self.module, "checkParameterValuesBeforeExecuting"):
                func = getattr(self.module, "checkParameterValuesBeforeExecuting")
                return func(self, grass_parameters, context)
        return super().checkParameterValues(grass_parameters, context)



"""
***************************************************************************
    grass_plugin.py
    ---------------------
    Date                 : June 2021
    Copyright            : (C) 2021 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "June 2021"
__copyright__ = "(C) 2021, Alexander Bruy"


from qgis.core import QgsApplication, QgsRuntimeProfiler

with QgsRuntimeProfiler.profile("Import GRASS Provider"):
    from grassprovider.grass_provider import GrassProvider


class GrassProviderPlugin:
    def __init__(self):
        self.provider = GrassProvider()

    def initProcessing(self):
        QgsApplication.processingRegistry().addProvider(self.provider)

    def initGui(self):
        self.initProcessing()

    def unload(self):
        QgsApplication.processingRegistry().removeProvider(self.provider)
